# -*- coding: utf-8 -*-
from __future__ import print_function, division
import re
from PyAstronomy.pyaC import pyaErrors as PE
import pickle
import os
import uuid
import six
import six.moves as smo


def equal(dependsOn):
    return dependsOn


class Params:
    r"""
    Manage a set of parameters.

    This class provides a framework for managing parameter values
    and several associated aspects such as restrictions and relation.
    Parameter management is implemented in view of later use by a
    a fit routine.

    Parameters
    ----------
    paramNames : list of strings
        Specifies the names of the parameters to be managed.


    Attributes
    ----------
    __params : dict
        A dictionary containing entries of the kind [variable-name:value].
        The `__params` dictionary may only be access by the
        assignValue function to prevent causing inconsistencies,
        which may occur especially if relations exist.                     .
    paramNum : dict
        A dictionary assigning every parameter name to a number
        (e.g., paramNum[2] = "XYZ"). Such a numbering is important to
        guarantee the correct order of variables in
        the fitting process. The ordering is the same as the order in
        which the constructor obtains the parameter names.
    isFree : dict
        A dictionary associating each parameter with a boolean saying
        whether it shall be considered a free parameter during
        optimization.
    isRestricted : dict
        A dictionary associating each parameter with a boolean
        saying whether there is a restriction to the allowed
        range (see `restrictions`).
    restrictions : dict
        A dictionary containing entries of the form
        [parName:[lower, upper]. Here 'lower' and 'upper' are floats
        defining lower and upper limits for the variable's value.
    relations : dict
        Parameters may be interrelated, e.g., A = B + C.
        This dictionary holds the definition of such relations.
        In particular, the entries have the form
        {"b":[ ["a", func, ["b", "c"]], [...] ]...}.
        This means that a = func(b, c). func is
        a function pointer. Whenever b is assigned a new value,
        'a' has to be updated
        as well. Note that it is important that the independent
        variables "know" about the relation, because if their value
        changes, the value of the dependent
        variable (in this case `a`) has to be updated.
    conditionalRestrictions : dict
        A dictionary holding the `conditional restrictions', i.e.,
        complex restrictions, which may, e.g., depend in the values
        of other parameters. The dictionary key is a unique ID
        generated, when a conditional restriction is added. For each
        key, the dictionary holds a tuple of which the first entry
        is a list holding the names of the parameters on which the
        conditional restrictions depends and the second is a
        callable, which is called with the values of those parameters
        specified in the first entry. The callable must return a
        float that specifies the penalty (or reward) depending on
        the given parameter values. Because conditional restrictions
        are referred to using a unique ID, their name (i.e., ID) does
        not change if models are combined.

    Notes
    -----

    Different models make different demands on the variables.
    For example, only certain ranges may be valid, some are constant
    and others not, or there may be a functional dependence between
    different variables. This class provides a framework to
    manage a parameter set appropriately.

    Depending on what kind of model is currently adapted, the number,
    names, allowed ranges, and interdependencies of variables can differ.
    On creation, this class is given a list with parameter names
    to manage. Those can than be assigned values. Parameters can be
    "thawed", i.e., regarded free during the fit process, or
    frozen. The allowed ranges can be restricted either on one or
    both sides, and interdependencies can be defined via the `relate`
    method.
    """

    def __add__(self, right):
        r"""
        This member allows to combine the properties of two Param class \
        instances. All parameter properties will be copied into the new \
        Parameter class instance, which is returned by this function.

        .. caution:: All parameter names must be unique!
        """
        paLeft = self.parameters()
        paRight = right.parameters()
        # Check whether parameters have unique names
        for p in paLeft:
            if p in paRight:
                raise (PE.PyANameClash(
                    "At least two parameters share the same name.", where="Params::__add__"))
        # List of new names
        nlist = list(paLeft.keys())
        nlist.extend(paRight.keys())
        # Create new parameter objects
        result = Params(nlist)
        # Combine properties
        result.paramNum = self.paramNum.copy()
        pn = len(result.paramNum)
        for i in smo.range(len(right.paramNum)):
            result.paramNum[pn + i] = right.paramNum[i]
        result.isFree = self.isFree.copy()
        result.isFree.update(right.isFree)
        result.isRestricted = self.isRestricted.copy()
        result.isRestricted.update(right.isRestricted)
        result.restrictions = self.restrictions.copy()
        result.restrictions.update(right.restrictions)
        result.relations = self.relations.copy()
        result.relations.update(right.relations)
        result.conditionalRestrictions = self.conditionalRestrictions.copy()
        result.conditionalRestrictions.update(right.conditionalRestrictions)
        result.assignValue(self.parameters())
        result.assignValue(right.parameters())
        return result

    def renameParameter(self, old, new):
        r"""
        Rename an existing parameter.

        Parameters
        ----------
        old : string
            The existing (old) name.
        new : string
            The new name.
        """
        self.__checkForParam(old)
        if new in self.__params:
            raise (PE.PyAValError("Parameter already exists: " +
                                  new, where="Params::renameParameter"))
        self.__params[new] = self.__params[old]
        del self.__params[old]
        for i in smo.range(len(self.paramNum)):
            if self.paramNum[i] == old:
                self.paramNum[i] = new
                break
        self.isFree[new] = self.isFree[old]
        del self.isFree[old]
        self.isRestricted[new] = self.isRestricted[old]
        del self.isRestricted[old]
        self.restrictions[new] = self.restrictions[old]
        del self.restrictions[old]
        self.relations[new] = self.relations[old]
        del self.relations[old]
        # Loop through relations, search, and replace occurrences of the old name.
        for p in six.iterkeys(self.__params):
            relations = self.relations[p]
            if relations == []:
                continue
            for k in smo.range(len(relations)):
                relat = relations[k]
                if relat[0] == old:
                    relat[0] = new
                for i in smo.range(len(relat[2])):
                    if relat[2][i] == old:
                        relat[2][i] = new
                self.relations[p][k] = relat
        # Loop over conditional restrictions and replace occurrences of the old names
        for name, v in six.iteritems(self.conditionalRestrictions):
            for i, p in enumerate(v[0]):
                # Loop over input-parameter names and replace if necessary
                if p == old:
                    # Needs to be replaced
                    self.conditionalRestrictions[name][0][i] = new

    def __init__(self, paramNames):
        i = 0
        # Define instance properties
        self.__params = {}
        self.paramNum = {}
        self.isFree = {}
        self.isRestricted = {}
        self.restrictions = {}
        self.relations = {}
        self.conditionalRestrictions = {}
        for n in paramNames:
            self.__params[n] = 0.0
            self.isFree[n] = False
            self.paramNum[i] = n
            i += 1
            # Firstly, the lower bound, secondly, the upper bound
            self.isRestricted[n] = [False, False]
            self.restrictions[n] = [None, None]
            self.relations[n] = []

    def addConditionalRestriction(self, pars, func):
        r"""
        Define a conditional restriction.

        Conditional restrictions can be used to modify the
        behavior in a more complex manner. For instance,
        penalties can be added to the objective function
        depending on the relation of one or more parameters.

        The given function is evaluated in each iteration
        and its return value (a float) is added to the
        objective function (e.g., chi square).

        Parameters
        ----------
        pars : list of strings
            The names of the parameters the given function
            needs as input.
        func : callable object
            This callable object must take the specified
            parameters (in that exact order) as input. It
            must return a float, which is added to the
            value of the objective function.

        Returns
        -------
        identifier : string
            A unique ID used to refer to the conditional
            restriction.
        """
        for p in pars:
            self.__checkForParam(p)
        name = uuid.uuid4()
        self.conditionalRestrictions[name] = (pars[:], func)
        return name

    def removeConditionalRestriction(self, id):
        r"""
        Remove an existing conditional constraint.

        Parameters
        ----------
        id : string
            The identifier used to refer to the conditional
            constraint (returned by `addConditionalRestriction`).
        """
        if not id in self.conditionalRestrictions:
            raise (PE.PyAValError("No conditional restriction with ID '" + str(id)))
        del self.conditionalRestrictions[id]

    def showConditionalRestrictions(self, toScreen=True):
        r"""
        Show conditional restrictions.

        Parameters
        ----------
        toScreen : boolean, optional
            If True (default), the output is written to stdout.

        Returns
        -------
        Output : list of strings
            The output as a list of strings.
        """
        ll = []
        nc = 80
        ll.append("-" * nc)
        ll.append("    Conditional restrictions")
        ll.append("-" * nc)
        for name, v in six.iteritems(self.conditionalRestrictions):
            s = "ID: " + str(name) + ", parameters: "
            s += ', '.join(v[0])
            ll.append(s)
        ll.append("-" * nc)
        if toScreen:
            for l in ll:
                print(l)
        return ll

    def applyConditionalRestrictions(self, fullout=False):
        r"""
        Apply all conditional restrictions.

        Parameters
        ----------
        fullout : boolean, optional
            If True, a dictionary holding the values
            of the individual restrictions is returned.
            The IDs are used as dictionary keys. The
            default is False.

        Returns
        -------
        Modification : float
            The summed value of the existing conditional restrictions.
        Individual values : dictionary, optional
            The contributions of the individual conditional
            restrictions.
        """
        result = 0.0
        if fullout:
            fo = {}
        for name, v in six.iteritems(self.conditionalRestrictions):
            vs = ()
            for p in v[0]:
                vs += (self[p],)
            val = v[1](*vs)
            if val is None:
                raise (PE.PyAValError("Conditional restriction with ID '" + str(name) + "' returned None.",
                                      solution="Did you forget to return 0.0, in case the condition is not fulfilled?"))
            if fullout:
                fo[name] = val
            result += val
        if fullout:
            return result, fo
        else:
            return result

    def __toList(self, x):
        r"""
        Return a list of x is a string and do nothing of x is already a list.
        """
        if isinstance(x, six.string_types):
            return [x]
        return x

    def __checkForParam(self, name):
        r"""
        Checks whether parameter exists and throws an exception if this is not the case.
        No return value.
        """
        if not self.hasParam(name):
            raise (PE.PyAValError("No such parameter: \"" + name + "\".\n  Available parameters: " +
                                  ', '.join(list(self.parameters().keys())), where="Params::__checkForParam"))

    def hasParam(self, name):
        r"""
        Check whether parameter exists.

        Parameters
        ----------
        name : string
            The parameter name.

        Returns
        -------
        flag : boolean
            True if parameter exists, False otherwise.
        """
        if not name in self.__params:
            return False
        return True

    def availableParameters(self):
        r"""
        Provides a list of existing parameters.

        Returns
        -------
        Parameters : list of strings 
            A list with the names of available parameters.
        """
        return list(self.__params.keys())

    def parameters(self):
        r"""
        Obtain parameter names and values.

        Returns
        -------
        Name-value : dict
            A dictionary with the names and values of all parameters
            ({"parName":value, ...}).
        """
        return self.__params.copy()

    def __getitem__(self, name):
        r"""
        Access parameter values.

        Parameters
        ----------
        name : string
            Parameter name.

        Returns
        -------
        Value : float
            The value of the parameter specified by `name`.
        """
        self.__checkForParam(name)
        return self.__params[name]

    def assignValue(self, namval):
        r"""
        Define new parameter values.

        Parameters
        ----------
        namval : dict
            A dictionary containing ['name':'value'] pairs.
        """
        for n, v in six.iteritems(namval):
            self.__checkForParam(n)
            self.__params[n] = v
            # Check for relations
            if len(self.relations[n]) > 0:
                # Yes, there are relations
                for tie in self.relations[n]:
                    arg = []
                    if len(tie[2]) == 1:
                        # Function takes a single argument
                        # Get value of dependent variable
                        arg = [self[tie[2][0]]]
                    else:
                        # Function takes a list of arguments
                        for name in tie[2]:
                            arg.append(self[name])
                    # Assign value to dependent variable
                    self.assignValue({tie[0]: tie[1](*arg)})

    def thaw(self, name):
        r"""
        Thaw (regard as free) a parameter.

        Parameters
        ----------
        name : string or list of strings
            The name(s) of the parameter(s) to be thawed.
        """
        names = self.__toList(name)
        for n in names:
            # Check whether the parameter is a dependent variable in
            # a relation
            for pn, rels in six.iteritems(self.relations):
                for rel in rels:
                    if rel[0] == n:
                        raise (PE.PyAParameterConflict("You tried to free the parameter '" + n +
                                                       "', which is a dependent variable in a relation.",
                                                       solution="Use 'untie' first to remove the relation."))
            self.__checkForParam(n)
            self.isFree[n] = True

    def freeze(self, name):
        r"""
        Freeze parameter(s) (contrary of thaw).

        Parameters
        ----------
        name : string or list of strings
            The name(s) of the parameter(s) to be frozen.
        """
        names = self.__toList(name)
        for n in names:
            self.__checkForParam(n)
            self.isFree[n] = False

    def freeParameters(self):
        r"""
        Get names and values of free parameters.

        Returns
        -------
        Free parameters: dict
             Dictionary containing the names and values
             of all free parameters ({"parName":value, ...}).
        """
        tuplist = []
        for i in smo.range(len(self.paramNum)):
            name = self.paramNum[i]
            if self.isFree[name]:
                tuplist.append((name, self.__params[name]))
        result = dict(tuplist)
        return result

    def frozenParameters(self):
        r"""
        Get names and values of frozen parameters.

        Returns
        -------
        Frozen parameters: dict
             Dictionary containing the names and values
             of all frozen parameters ({"parName":value, ...}).
        """
        tuplist = []
        for i in smo.range(len(self.paramNum)):
            name = self.paramNum[i]
            if not self.isFree[name]:
                tuplist.append((name, self.__params[name]))
        result = dict(tuplist)
        return result

    def numberOfFreeParams(self):
        r"""
        Get number of free parameters.

        Returns
        -------
        n : int
            The number of free parameters (determined by `isFree`).
        """
        c = 0
        for v in six.itervalues(self.isFree):
            if v:
                c += 1
        return c

    def freeParamNames(self):
        r"""
        Get the names of the free parameters.

        Returns
        -------
        Free parameters : list of strings
            The names of the free parameters.
            The order is determined by the
            `paramNum` attribute.
        """
        result = []
        for i in smo.range(len(self.paramNum)):
            name = self.paramNum[i]
            if self.isFree[name]:
                result.append(name)
        return result

    def setFreeParams(self, X):
        r"""
        Change the values of the free parameters.

        Parameters
        ----------
        X : list of floats
            Contains the values for all free parameters. Note
            that the numbering is according to the `paramNum` attribute.

        Notes
        -----
        This method is primarily implemented to be used by fit routines.
        """
        if len(X) != self.numberOfFreeParams():
            raise (PE.PyAValError(
                "Number of supplied parameters does not match number of free parameters.", where="Params::setFreeParams"))
        c = 0
        for i in smo.range(len(self.__params)):
            name = self.paramNum[i]
            if self.isFree[name]:
                self.assignValue({name: X[c]})
                c += 1

    def getFreeParams(self):
        r"""
        Get values of free parameters.

        Returns
        -------
        Values : list of floats
            The values of the free parameters. Note that the order is
            determined by the `paramNum` attribute.
        """
        result = []
        for i in smo.range(len(self.__params)):
            name = self.paramNum[i]
            if self.isFree[name]:
                result.append(self.__params[name])
        return result

    def setRestriction(self, restricts):
        r"""
        Apply restrictions to parameter ranges.

        Parameters
        ----------
        restricts : dict
            A dictionary associating name and [lower-bound, upper-bound].
            If no boundary shall exist on one side, use 'None'.
        """
        for name, v in six.iteritems(restricts):
            self.__checkForParam(name)
            if (v[0] is not None) and (v[1] is not None):
                if v[0] >= v[1]:
                    raise (PE.PyAValError(
                        "Lower bound < upper bound must be fulfilled!", where="Params::setRestriction"))
            self.isRestricted[name] = [(v[0] is not None), (v[1] is not None)]
            self.restrictions[name] = v

    def delRestriction(self, parName):
        r"""
        Delete restriction

        Parameters
        ----------
        parName : string
            Name of restricted parameter
        """
        self.__checkForParam(parName)
        self.isRestricted[parName] = [False, False]
        self.restrictions[parName] = [None, None]

    def getRestrictions(self):
        r"""
        Get all restrictions.

        Returns
        -------
        Restrictions : dict
            Dictionary associating parameter name and restriction
            (see `restrictions` attribute).
        """
        return self.restrictions

    def untie(self, parName, forceFree=False):
        r"""
        Remove all relations of parameter "parName".

        After this operation, the parameter no longer depends 
        on other parameters. Unless foreFree is True, the parameter
        referred to by "parName" is "frozen".

        Parameters
        ----------
        parName : string
            The name of the dependent variable, which shall be untied.
        forceFree : boolean
            Set parName to "free" instead of "frozen" if set to True.
        """
        for p in six.iterkeys(self.relations):
            self.relations[p] = [
                r for r in self.relations[p] if r[0] != parName]
        self.isFree[parName] = forceFree

    def relate(self, parName1, pars, func=None, force=False):
        r"""
        Apply functional relation between parameters.

        Relates parameters, i.e., par1 = func(pars). The two values are related by 'func'. \
        In order to be tied, all involved parameters have to be free.

        Parameters
        ----------
        parName1 : string
            The name of the dependent variable.
        pars : list of strings
            Names of independent variables.
        func : callable
            The function that determines the form of the dependence.
            The default is "equal".
        force : boolean, optional
            Set to True in order to suppress error when the dependent
            variable is frozen. 
        """
        self.__checkForParam(parName1)
        if not self.isFree[parName1] and not force:
            raise (PE.PyAValError(parName1 + " is not free.",
                                  where="Params::relate", solution="Use 'thaw' to free parameter."))
        if isinstance(pars, six.string_types):
            pars = [pars]
        for p in pars:
            self.__checkForParam(p)
        # Check whether the dependent variable is in the list of independent ones
        if parName1 in pars:
            raise (PE.PyAValError("The dependent variable (" + parName1 +
                                  ") cannot be in the list of independent variables."))
        # Use equal function if None is specified.
        if func is None:
            func = equal
        # Append to 'relations' list
        for p in pars:
            # If parameter 'p' is set, redefine 'parName1' using function 'func' with parameters 'pars'
            self.relations[p].append([parName1, func, pars])
        # Update parameter values (on re-assigned the values of the independent
        # parameters, the value of the dependent parameter will be updated)
        for p in pars:
            self.assignValue({p: self.__params[p]})
        self.isFree[parName1] = False

    def getPenalty(self, penaltyFact=1e20):
        r"""
        Get applied penalty for current parameter set.

        Parameters
        ----------
        penaltyFact : float, optional
            The higher the number the higher the penalty for small deviations
            (default is 10**20).

        Returns
        -------
        Penalty : float
            The applied penalty for current parameter set.
        Penalties : dict
            A dictionary with a key for every parameter values
            for which a bound is violated. The value is the
            amount by which the bound (upper or lower) is
            violated.
        """
        result = {}
        totval = 0.0
        for name, r in six.iteritems(self.isRestricted):
            if r[0]:
                # There is a lower bound
                if self.__params[name] < self.restrictions[name][0]:
                    result[name] = abs(self.__params[name] -
                                       self.restrictions[name][0])
                    totval += result[name]
                    # It cannot be smaller and greater...continue
                    continue
            if r[1]:
                if self.__params[name] > self.restrictions[name][1]:
                    result[name] = abs(self.__params[name] -
                                       self.restrictions[name][1])
                    totval += result[name]
                    continue
        totval *= penaltyFact
        return (totval, result)

    def getRelationsOf(self, parName):
        r"""
        Obtain relations for a parameter.

        Parameters
        ----------
        parName : string
            The name of the parameter of which relations shall be searched.

        Returns
        -------
        Relations : list of relations
            Those relations in which `parName` is the dependent variable, i.e.,
            parName = f(x,y).
        """
        self.__checkForParam(parName)
        result = []
        for k in six.iterkeys(self.__params):
            for r in self.relations[k]:
                if r[0] == parName:
                    if not r in result:
                        result.append(r)
        return result

    def saveState(self, fn=None, clobber=False):
        r"""
        Save the state of the fitting object.

        This method collects the parameter values, the applied restrictions,
        and the information whether parameters are free or frozen and saves
        them to the specified file (if given) using pickle.
        The saved state can be restored using
        the `restoreState` method.

        .. note:: Unfortunately, "relations" cannot be saved.

        Parameters
        ----------
        fn : string, optional
            The filename to which the state shall be written.
            If None, the output will not be written to a file.
        clobber : boolean, optional
            If True, existing files will be overwritten (default is False).

        Returns
        -------
        Saved data : dict
            The dictionary containing the data saved to the file.
        """
        dat = {}
        dat["parameters"] = self.parameters()
        dat["restrictions"] = self.getRestrictions()
        dat["frozenParameters"] = self.frozenParameters()
        if isinstance(fn, six.string_types):
            if os.path.isfile(fn) and (not clobber):
                raise (PE.PyANameClash("The file '" + fn + "' exists.", where="saveState",
                                       solution="Change filename or set clobber to True."))
            pickle.dump(dat, open(fn, 'wb'))
        return dat

    def restoreState(self, resource):
        r"""
        Restores parameter values from file or dictionary.

        Parameters
        ----------
        resource : string or dictionary
            If string, it is interpreted as filename of a pickle file holding the
            data dictionary. If dictionary, it uses the data saved in it; note that
            a valid data dictionary is returned by `saveState`.
        """
        if isinstance(resource, six.string_types):
            pd = pickle.load(open(resource, 'rb'))
        else:
            pd = resource
        # Free ALL values
        for k in six.iterkeys(self.parameters()):
            rels = self.getRelationsOf(k)
            if len(rels) == 0:
                self.thaw([k])
        # Assign parameter values
        self.assignValue(pd["parameters"])
        # Assign restrictions
        self.setRestriction(pd["restrictions"])
        # Freeze the frozen parameters
        self.freeze(pd["frozenParameters"])

    def parameterSummary(self, lines=None, toScreen=True, prefix="", onlyParams=False):
        r"""
        Writes a summary of the parameters in text form.

        Parameters
        ----------
        lines : list of strings, optional
            If given, the output will be attached to this list.
        toScreen : boolean, optional
            If False, screen output will be suppressed
            (default is True).
        prefix : string, optional
            A prefix applied to every output line (e.g., '#')

        @FIXME - This is antique...

        Attaches the text lines
        to the lines list if given. If lines is not given, 'toScreen' will be set to True
        and the result is written to the screen.
        Returns either a new list containing the text, or (if given) the 'lines' list
        with result appended.
        A 'prefix' may be specified (e.g., '#'), which will preceded every line.
        """
        if lines is None:
            lines = []
            toScreen = True

        # Find maximal length of parameter name
        plen = 0
        for k, v in six.iteritems(self.__params):
            if len(k) > plen:
                plen = len(k)

        # Comppars - dictionary mapping Component number -> list of variables
        compPars = {}
        for name in six.iterkeys(self.__params):
            r = re.match(".*_([0-9]+)", name)
            if r is not None:
                cn = int(r.group(1))
            else:
                cn = -1
            if cn in compPars:
                compPars[cn].append(name)
            else:
                compPars[cn] = [name]

        for k, v in six.iteritems(compPars):
            compPars[k] = sorted(v)

        components = sorted(compPars.keys())
        for com in components:
            if com != -1 and not onlyParams:
                lines.append(
                    prefix + "Parameters for component no. " + str(com))
                lines.append(prefix + "-------------------------------------")
            for i in smo.range(len(compPars[com])):
                #      name = self.paramNum[i]
                name = compPars[com][i]

                restricted = (
                    self.isRestricted[name][0] or self.isRestricted[name][1])
                line = (prefix + "Parameter: %" + str(plen) + "s, value:  %13.8g, free: %5s, restricted: %5s") % (
                    name, self.__params[name], str(self.isFree[name]), str(restricted))
                if restricted:
                    line += ", lower bound: "
                    if self.restrictions[name][0] is not None:
                        line += "%12.6f" % self.restrictions[name][0]
                    else:
                        line += "None"
                    line += ", " + "upper bound: "
                    if self.restrictions[name][1] is not None:
                        line += "%12.6f" % self.restrictions[name][1]
                    else:
                        line += "None"

                relas = self.getRelationsOf(name)
                if len(relas) > 0:
                    for r in relas:
                        line += ",  Relation: " + r[0] + " = f("
                        for k in range(len(r[2]) - 1):
                            line += r[2][k] + ", "
                        line += r[2][len(r[2]) - 1] + ")"
                lines.append(line)

        if toScreen:
            for l in lines:
                print(l)

        return lines



# -*- coding: utf-8 -*-
from __future__ import print_function, division
from numpy import mean, median, std
import numpy as np
import re
import os
from scipy.stats import spearmanr, pearsonr
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.funcFit.utils import ic, hpd, quantiles
import itertools
from PyAstronomy import pyaC as PC
import six
import six.moves as smo


try:
    import matplotlib.pylab as plt
except ImportError:
    pass


class TraceAnalysis2:
    r"""
    Support to analyze MCMC chains.

    This class provides a number of plotting methods. Note that
    **you still need to call *show()* from pylab** to see the
    result.

    Parameters
    ----------
    resource : string
        A filename of the form "*.emcee", assumed to be
        a trace produced by emcee. 

    Attributes
    ----------
    burn : int
        Applies a burn-in. All iterations
        earlier than `burn` will be neglected.

    thin : int
        Applies thinning to each chain.
        Retains every `k` th sample,
        where `k` is an integer value.  
    """

    def _parmCheck(self, parm):
        r"""
        Checks whether a trace is available for the given parameter.
        If not, it throws an exception.

        Parameters
        ----------
        parm : string
            Variable name.
        """
        if not parm in self.tracesDic:
            raise (PE.PyAValError("No trace available for parameter " + parm + ".\n  Available parameters: " + ', '.join(list(self.tracesDic.keys())),
                                  where="TraceAnalysis"))

    def __plotsizeHelper(self, size):
        r"""
        Helps to define the optimum plot size for large big-picture plots.
        """
        c = 1
        r = 1
        while c * r < size:
            c += 1
            if c * r >= size:
                break
            else:
                r += 1
        return c, r

    def _loadEMCEEChain(self, fn=None, burn=0):
        r"""
        Load chains from file or adjust parameters.

        Based on the loaded data base, burn-in and walker selection
        is applied.

        Parameters
        ----------
        fn : string, optional
            If given, the filename of the emcee chain file.
            If not given, the previously saved data base will
            be used.
        burn : int, optional
            Number of burn-in steps.
        """
        if not fn is None:
            self._emceedat = np.load(fn)
        self.emceepnames = list(
            self._emceedat["pnames"]) + ["lnpost", "lnprior", "lnl"]
        # Build stateDic
        self.stateDic = {"stochastics": dict(zip(list(self.emceepnames), [
                                             None] * len(list(self.emceepnames)))), "sampler": {}}

        # Flatten the chain
        s = self._emceedat["chain"].shape
        # s[0] = No. of chains, s[1] = Length of individual chain, s[2] = No. of parameters

        if not hasattr(self, "_selectedWalker"):
            # Select all walker ...
            selectedWalker = np.arange(s[0])
            # ... and save selection
            self._selectedWalker = selectedWalker
        # Use saved selection
        selectedWalker = self._selectedWalker

        # No. of walker considered in the analysis
        nchains = len(selectedWalker)

        # Check selected chains
        if np.any(selectedWalker > s[0]) or np.any(selectedWalker < 0):
            raise (PE.PyAValError("You selected at least one walker beyond the valid range (0 - " + str(s[0] - 1) + ").",
                                  solution="Adjust walker selection."))

        self.emceechain = np.zeros((nchains * (s[1] - burn), s[2] + 3))
        self.emceechain[::, 0:-3] = self._emceedat["chain"][selectedWalker,
                                                            burn:, ::].reshape(nchains * (s[1] - burn), s[2])

        def getprop(prop):
            c = self._emceedat[prop][selectedWalker, burn:]
            d = c.reshape(nchains * (s[1] - burn))
            return d

        # Incorporate posterior, prior, and likelihood
        self.emceechain[::, -1] = getprop("lnl")
        self.emceechain[::, -2] = getprop("lnprior")
        self.emceechain[::, -3] = getprop("lnpost")

        self.stateDic["sampler"]["_iter"] = self.emceechain.shape[0]
        self.stateDic["sampler"]["_burn"] = None
        self.stateDic["sampler"]["_thin"] = None

        # Populate tracesDic
        self.tracesDic = {
            p: self.emceechain[::, self.emceepnames.index(p)] for p in self.emceepnames}

    def _checkPackage(self, p):
        r"""
        Check whether package is availabel and raise exception otherwise.

        Parameters
        ----------
        p : string
            Name of package
        """
        if not ic.check[p]:
            raise (PE.PyARequiredImport("The package '" + str(p) + "' is not currently installed.",
                                        solution="Please install " + str(p)))

    def selectWalkers(self, ws):
        r"""
        Select walkers for emcee chains.

        Parameters
        ----------
        ws : list or array of integers
            The walker to be considered in the analysis. Counting
            starts at zero.
        """
        self._selectedWalker = np.array(ws, dtype=int)
        if self.dbtype == "emcee":
            # Apply burn-in to individual walkers
            self._loadEMCEEChain(burn=self.burn)
        else:
            raise (PE.PyAAlgorithmFailure(
                "Walkers can only be selected for emcee data bases. Current data base type: " + str(self.dbtype)))

    def selectedWalkers(self):
        r"""
        Get the list of selected walkers.

        Parameters
        ----------
        walkers : array
            The selected walkers.
        """
        if not hasattr(self, "_selectedWalker"):
            raise (PE.PyAValError(
                "Cannot determine the selected walkers. It appears that no emcee chain was loaded."))
        return self._selectedWalker[:]

    def numberOfWalkers(self):
        r"""
        Get number of walkers in emcee chain.

        Returns
        -------
        nWalkers : int
            Number of available walkers
        """
        if not hasattr(self, '_emceedat'):
            raise (PE.PyAValError(
                "Cannot determine the number of walkers. It appears that no emcee chain was loaded."))
        return self._emceedat["chain"].shape[0]

    def __init__(self, resource, db=None):
        if isinstance(resource, six.string_types):
            # Resource is a filename
            if not os.path.isfile(resource):
                raise (PE.PyAFileError(resource, "ne"))
            self.file = resource

            if not (re.match(r".*\.hdf5", resource) is None):
                db = "hdf5"
            if not (re.match(r".*\.zlib", resource) is None):
                db = "hdf5"
            if not (re.match(r".*\.emcee", resource) is None):
                db = "emcee"

            if db == "emcee":
                self.dbtype = "emcee"
                self._loadEMCEEChain(resource)
            else:
                raise (PE.PyAValError("Database type '" +
                                      db + "' is currently not supported."))

        # Set default burn-in and thinning
        self.burn = 0
        self.thin = 1

        # A set of chains by default belonging to a 'technical' set (not usually parameters)
        self.technicalChains = ["lnprior", "lnl", "lnpost", "deviance"]

    def __getitem__(self, parm):
        r"""
        Returns the trace for parameter `parm`.

        Parameters
        ----------
        parm : string
            Name of the parameter.

        Returns
        -------
        trace : array
            The trace for the parameter.
        """
        self._parmCheck(parm)
        if self.dbtype == "emcee":
            index = self.emceepnames.index(parm)
            # Burn-in is handled during loading of chain
            return self.emceechain[0::self.thin, index]

    def __str__(self):
        r"""
        Prints basic information on the current MCMC sample file.
        """
        info = "MCMC database - Basic information:\n"
        info += "----------------------------------\n\n"
        if hasattr(self, "file"):
            info += "MCMC sample file: " + self.file + "\n\n"
        info += "  Stochastics:  " + \
            str(list(self.stateDic["stochastics"].keys())) + "\n\n"
        info += "  Sampler:      " + "Iterations: " + \
            str(self.stateDic["sampler"]["_iter"]) + "\n"
        info += "                " + "Burn:       " + \
            str(self.stateDic["sampler"]["_burn"]) + "\n"
        info += "                " + "Thin:       " + \
            str(self.stateDic["sampler"]["_thin"]) + "\n\n"
        info += "  Note: There is more information available using the state() method.\n"
        return info

    def availableParameters(self, wtech=False):
        r"""
        Returns list of available parameter names.

        Parameters
        ----------
        wtech : boolean, optional
            If True, include 'technical' traces (e.g., deviance and likelihood)
            if available.

        Returns
        -------
        Parameters : list of strings
            Available parameter traces
        """
        if not wtech:
            return [p for p in self.availableParameters(True) if p not in self.technicalChains]
        else:
            return list(self.stateDic["stochastics"].keys())

    def availableTraces(self):
        r"""
        Returns a list of available PyMC *Trace* objects
        """
        return list(self.tracesDic.keys())

    def state(self):
        r"""
        Returns dictionary containing basic information
        on the sampling process.
        """
        return self.stateDic

    def plotTrace(self, parm, fmt='b-'):
        r"""
        Plots the trace.

        Parameters
        ----------
        parm : string
            The variable name.
        fmt : string, optional
            The matplotlib format string used to plot the trace.
            Default is 'b-'.
        """
        if not ic.check["matplotlib"]:
            PE.warn(PE.PyARequiredImport("To use 'plotTrace', matplotlib has to be installed.",
                                         solution="Install matplotlib."))
            return
        self._parmCheck(parm)
        plt.plot(self[parm], fmt, label=parm + " trace")
        plt.legend()

    def plotTraceHist(self, parm):
        r"""
        Plots trace and histogram (distribution).

        Parameters
        ----------
        parm : string
            The variable name.
        """
        if not ic.check["matplotlib"]:
            PE.warn(PE.PyARequiredImport("To use 'plotTraceHist', matplotlib has to be installed.",
                                         solution="Install matplotlib."))
            return
        self._parmCheck(parm)

        plt.subplot(1, 2, 1)
        self.plotTrace(parm)
        plt.subplot(1, 2, 2)
        self.plotHist(parm)

    def plotHist(self, parsList=None):
        r"""
        Plots distributions for a number of traces.

        Parameters
        ----------
        parsList : string or list of strings, optional,
            Refers to a parameter name or a list of parameter names.
            If None, all available parameters are plotted.
        """
        if not ic.check["matplotlib"]:
            PE.warn(PE.PyARequiredImport("To use 'plotHists', matplotlib has to be installed.",
                                         solution="Install matplotlib."))
            return
        if isinstance(parsList, six.string_types):
            parsList = [parsList]
        tracesDic = {}
        if parsList is not None:
            for parm in parsList:
                self._parmCheck(parm)
                tracesDic[parm] = self[parm]
        else:
            # Use all available traces
            for parm in self.availableParameters():
                tracesDic[parm] = self[parm]

        cols, rows = self.__plotsizeHelper(len(tracesDic))

        for i, [pars, trace] in enumerate(tracesDic.items()):
            if len(tracesDic) > 1:
                plt.subplot(rows, cols, i + 1)
            plt.hist(trace, label=pars + " hist")
            plt.legend()

    def plotDeviance(self, parsList=None):
        r"""
        Plots value of deviance over parameter values encountered during sampling.

        Parameters
        ----------
        parsList : string or list of strings, optional,
            Refers to a parameter name or a list of parameter names.
            If None, all available parameters are plotted.
        """
        if not ic.check["matplotlib"]:
            PE.warn(PE.PyARequiredImport("To use 'plotHists', matplotlib has to be installed.",
                                         solution="Install matplotlib."))
            return
        if isinstance(parsList, six.string_types):
            parsList = [parsList]
        tracesDic = {}
        if parsList is not None:
            for parm in parsList:
                self._parmCheck(parm)
                tracesDic[parm] = self[parm]
        else:
            # Use all available traces
            for parm in self.availableParameters():
                tracesDic[parm] = self[parm]

        ps = self.__plotsizeHelper(len(tracesDic))

        for i, [pars, trace] in enumerate(six.iteritems(tracesDic), 1):
            plt.subplot(ps[0], ps[1], i)
            plt.xlabel(pars)
            plt.ylabel("Deviance")
            plt.plot(self[pars], self["deviance"], '.')

    def plotHists(self, parsList=None, **histArgs):
        print("TraceAnalysis::plotHists() - Warning! This function is deprecated. Use plotHist() instead.")
        self.plotHist(parsList, **histArgs)

    def hpd(self, parm, trace=None, cred=0.95):
        r"""
        Calculates highest probability density interval (HPD, minimum width BCI).

        Parameters
        ----------
        parm : string
            Name of parameter
        cred : float, optional
            Credibility level. Defaults to 0.95, i.e., the 95\% HPD will
            be calculated.
        trace : array, optional
            If a trace is given, it will be used in the calculation instead of the
            trace for `parm` stored in the class. Note that the parm will be
            ignored in this case!

        Returns
        -------
        HPD : tuple
            The lower and upper bound of the credibility interval.
        """
        if trace is None:
            self._parmCheck(parm)
            return hpd(self[parm], cred)
        else:
            return hpd(trace, cred)

    def quantiles(self, parm, qlist=None):
        r"""
        Quantiles for given trace.

        Parameters
        ----------
        parm : string
            Name of parameter
        qlist : list of floats (0-100), optional
            Specifies which quantiles shall be calculated. The default is
            2.5, 25, 50, 75, and 97.5 percent.

        Returns
        -------
        Quantiles : dictionary
            For each quantile (in percent) the corresponding value.
        """
        if qlist is None:
            qlist = [2.5, 25, 50, 75, 97.5]
        return quantiles(self[parm], qlist)

    def plotCorr(self, parsList=None, **plotArgs):
        r"""
        Produces correlation plots.

        Parameters
        ----------
        parsList : list of string,  optional
            If not given, all available traces are used.
            Otherwise a list of at least two parameters
            has to be specified.
        plotArgs : dict, optional
            Keyword arguments handed to plot procedure of
            pylab.
        """
        if not ic.check["matplotlib"]:
            PE.warn(PE.PyARequiredImport("To use 'plotCorr', matplotlib has to be installed.",
                                         solution="Install matplotlib."))
            return
        tracesDic = {}
        if parsList is not None:
            for parm in parsList:
                self._parmCheck(parm)
                tracesDic[parm] = self[parm]
            if len(tracesDic) < 2:
                raise (PE.PyAValError("For plotting correlations, at least two valid parameters are needed.",
                                      where="TraceAnalysis::plotCorr"))
        else:
            # Use all available traces
            for parm in self.availableParameters():
                tracesDic[parm] = self[parm]

        pars = list(tracesDic.keys())
        traces = list(tracesDic.values())

        fontmap = {1: 10, 2: 8, 3: 6, 4: 5, 5: 4}
        if not len(tracesDic) - 1 in fontmap:
            fontmap[len(tracesDic) - 1] = 3

        k = 1
        for j in range(len(tracesDic)):
            for i in range(len(tracesDic)):
                if i > j:
                    plt.subplot(len(tracesDic) - 1, len(tracesDic) - 1, k)
                    plt.title("Pearson's R: %1.5f" % self.pearsonr(
                        pars[j], pars[i])[0], fontsize='x-small')
                    plt.xlabel(pars[j], fontsize='x-small')
                    plt.ylabel(pars[i], fontsize='x-small')
                    tlabels = plt.gca().get_xticklabels()
                    plt.setp(tlabels, 'fontsize', fontmap[len(tracesDic) - 1])
                    tlabels = plt.gca().get_yticklabels()
                    plt.setp(tlabels, 'fontsize', fontmap[len(tracesDic) - 1])
                    plt.plot(traces[j], traces[i], '.', **plotArgs)
                if i != j:
                    k += 1

    def __hist2d(self, x, y, contour=False, bins=(200, 200), cmap="Purples", interpolation='nearest', origin="lower", colors="k"):
        r"""
        Parameters
        ----------
        bins : tuple of two ints
        """
        H, xedges, yedges = np.histogram2d(x, y, bins)
        extent = [xedges.min(), xedges.max(), yedges.min(), yedges.max()]
        if not contour:
            plt.imshow(H, extent=extent, interpolation=interpolation,
                       origin=origin, cmap=cmap, aspect="auto")
#      plt.colorbar(fontsize=4)
        else:
            CS = plt.contour(H, extent=extent, origin=origin, colors=colors)
            plt.clabel(CS, inline=1, fontsize=10)

    def correlationTable(self, parsList=None, coeff="pearson", noPrint=False):
        r"""
        Calculate and show parameter correlations

        Parameters
        ----------
        parsList : list of strings, optional
            A list of parameters for which to calculate the
            correlation. If None, all available parameters
            will be used.
        coeff : string, {"pearson", "spearman"}, optional
            The coefficient to be used. By default, Pearson's
            correlation coefficient will be used.
        noPrint : boolean, optional
            If True, the table output will be suppressed.

        Returns
        -------
        Coefficients : dictionary
            Maps each tuple of two parameter names to the associated
            correlation coefficient.
        """
        if parsList is None:
            pars = self.availableParameters()
        else:
            pars = parsList[:]
        if len(pars) < 1:
            raise (PE.PyAValError("You need to specify at least one parameter!"))

        if coeff == "pearson":
            coeFunc = self.pearsonr
        elif coeff == "spearman":
            coeFunc = self.spearmanr
        else:
            raise (PE.PyAValError("Unknown coefficient '" + str(coeff) + "'.",
                                  solution="Use either of 'pearson' or 'spearman'."))

        corrs = {}
        cc = itertools.combinations_with_replacement(pars, 2)
        for c in cc:
            corrs[c] = coeFunc(c[0], c[1])[0]

        if not noPrint:
            # Find width of columns
            colWidth = {}
            for p in pars:
                colWidth[p] = max(len(p), 9)
            head = (" " * (max(colWidth.values()) + 1)) + "|"
            for p in pars:
                head += (" %" + str(colWidth[p]) + "s |") % p
            print(head)
            print("-" * len(head))
            for p2 in pars:
                line = ("%" + str(max(colWidth.values())) + "s |") % (p2)
                for p1 in pars:
                    if (p2, p1) in corrs:
                        coe = corrs[(p2, p1)]
                    else:
                        coe = corrs[(p1, p2)]
                    line += " % 8.6f |" % coe
                print(line)
            print("-" * len(head))

        return corrs

    def plotCorrEnh(self, parsList=None, **plotArgs):
        r"""
        Produces enhanced correlation plots.

        Parameters
        ----------
        parsList : list of string,  optional
            If not given, all available traces are used.
            Otherwise a list of at least two parameters
            has to be specified.
        plotArgs : dict, optional
            Keyword arguments handed to plot procedures of
            pylab. The following keywords are available:
            contour,bins,cmap,origin,interpolation,colors
        """
        if not ic.check["matplotlib"]:
            PE.warn(PE.PyARequiredImport("To use 'plotCorr', matplotlib has to be installed.",
                                         solution="Install matplotlib."))
            return
        tracesDic = {}
        if parsList is not None:
            for parm in parsList:
                self._parmCheck(parm)
                tracesDic[parm] = self[parm]
            if len(tracesDic) < 2:
                raise (PE.PyAValError("For plotting correlations, at least two valid parameters are needed.",
                                      where="TraceAnalysis::plotCorr"))
        else:
            # Use all available traces
            for parm in self.availableParameters():
                tracesDic[parm] = self[parm]

        pars = list(tracesDic.keys())
        traces = list(tracesDic.values())

        fontmap = {1: 10, 2: 9, 3: 8, 4: 8, 5: 8}
        if not len(tracesDic) - 1 in fontmap:
            fontmap[len(tracesDic) - 1] = 8

        k = 1
        for j in range(len(tracesDic)):
            for i in range(len(tracesDic)):
                if i > j:
                    plt.subplot(len(tracesDic) - 1, len(tracesDic) - 1, k)
#          plt.title("Pearson's R: %1.5f" % self.pearsonr(pars[j],pars[i])[0], fontsize='x-small')
                    plt.xlabel(pars[j], fontsize=fontmap[len(tracesDic) - 1])
                    plt.ylabel(pars[i], fontsize=fontmap[len(tracesDic) - 1])
                    tlabels = plt.gca().get_xticklabels()
                    plt.setp(tlabels, 'fontsize', fontmap[len(tracesDic) - 1])
                    tlabels = plt.gca().get_yticklabels()
                    plt.setp(tlabels, 'fontsize', fontmap[len(tracesDic) - 1])
#          plt.plot(traces[j],traces[i],'.',**plotArgs)
                    self.__hist2d(traces[j], traces[i], **plotArgs)
                if i != j:
                    k += 1

    def correlationMatrix(self, toScreen=True, method="pearson", parList=None, covariance=False):
        r"""
        Calculates the correlation or covariance matrix.

        Parameters
        ----------
        parList : list of strings, optional
            The list of parameters used in the calculation.
            If not given, all available parameters will be
            used.
        toScreen : boolean, optional
            If True, the result will be printed to stdout
        method : string, {"pearson", "spearman"}
            The correlation coefficient to be used.
        covariance : boolean, optional
            If True, the covariance will be returned instead
            of the correlation. The default is False.

        Returns
        -------
        Parlist : list
            Parameter names in the order used in the calculation.
        Correlation matrix : 2d array
            The correlation matrix
        lines : list of strings
            Formatted version of the correlation matrix in the form
            of a list of strings.
        """
        if parList is None:
            parList = self.availableParameters()
        corFunc = None
        if method == "pearson":
            corFunc = self.pearsonr
        if method == "spearman":
            corFunc = self.spearmanr
        if corFunc is None:
            raise (PE.PyAValError("The method " + str(method) + " is currently not supported.",
                                  solution="Change method argument e.g. to 'pearson'."))
        for p in parList:
            self._parmCheck(p)
        if covariance:
            # The covariance is requested. In this case, the
            # correlation coefficient has to be multiplied by the
            # standard deviation(s).
            stds = {}
            for p in parList:
                stds[p] = self.std(p)
        # Calculate the matrix
        n = len(parList)
        matrix = np.zeros((n, n))
        for i in smo.range(n):
            for j in smo.range(n):
                matrix[i, j] = corFunc(parList[i], parList[j])[0]
                if covariance:
                    matrix[i, j] *= (stds[parList[i]] * stds[parList[j]])

        # Format the output
        lines = PC.matrix2doutput(
            matrix, colNames=parList, rowNames=parList, toScreen=toScreen)

        return parList, matrix, lines

    def pearsonr(self, parm1, parm2):
        r"""
        Calculates a Pearson correlation coefficient and the
        p-value for testing non-correlation.

        Parameters
        ----------
        parm1, parm2 : string
            The names of the two parameters used in the evaluation. 

        Returns
        -------
        Pearson correlation coefficient : float
        p-value : float  

        Notes
        -----
        Uses SciPy's *scipy.stats.pearsonr* to evaluate.

        The SciPy documentation of scipy.stats.pearsonr:

          The Pearson correlation coefficient measures the linear
          relationship between two data sets. Strictly speaking, Pearson's
          correlation requires that each data set be normally distributed. 
          Like other correlation coefficients, this one varies between
          -1 and +1 with 0 implying no correlation. Correlations of
          -1 or +1 imply an exact linear relationship. Positive
          correlations imply that as x increases, so does y. Negative
          correlations imply that as x increases, y decreases.
          The p-value roughly indicates the probability of an uncorrelated
          system producing data sets that have a Pearson correlation at
          least as extreme as the one computed from these data sets.
          The p-values are not entirely reliable but are probably reasonable
          for data sets larger than 500 or so.
        """
        self._parmCheck(parm1)
        self._parmCheck(parm2)
        return pearsonr(self[parm1], self[parm2])

    def spearmanr(self, parm1, parm2):
        r"""
        Calculates a Spearman rank-order correlation coefficient
        and the p-value to test for non-correlation.

        Parameters
        ----------
        parm1, parm2 : string
            The names of the two parameters used in the evaluation.

        Returns
        -------
        Spearman rank-order correlation coefficient : float
        p-value : float

        Notes
        -----
        Uses SciPy's *scipy.stats.spearmanr* to evaluate.

        The SciPy documentation of scipy.stats.spearmanr:

          The Spearman correlation is a nonparametric measure of
          the monotonicity of the relationship
          between two data sets. Unlike the Pearson correlation,
          the Spearman correlation does not assume that both data
          sets are normally distributed. Like other correlation coefficients,
          this one varies between -1 and +1 with 0 implying no correlation.
          Correlations of -1 or +1 imply an exact monotonic relationship.
          Positive correlations imply that as x increases, so
          does y. Negative correlations imply that as x increases,
          y decreases. The p-value roughly indicates the probability of
          an uncorrelated system producing data sets that have a Spearman
          correlation at least as extreme as the one computed from these
          data sets. The p-values are not entirely reliable but are
          probably reasonable for data sets larger than 500 or so.
        """
        self._parmCheck(parm1)
        self._parmCheck(parm2)
        return spearmanr(self[parm1], self[parm2])

    def mean(self, parm):
        r"""
        Calculate mean.

        Parameters
        ----------
        parm : string
            Name of parameter.

        Returns
        -------
          The mean : float
        """
        self._parmCheck(parm)
        return mean(self[parm])

    def median(self, parm):
        r"""
        Calculate median.

        Parameters
        ----------
        parm : string
            Name of parameter.

        Returns
        -------
          The median : float
        """
        self._parmCheck(parm)
        return median(self[parm])

    def std(self, parm):
        r"""
        Calculate standard deviation.

        Parameters
        ----------
        parm : string
            Name of parameter.

        Returns
        -------
          The standard deviation : float
        """
        self._parmCheck(parm)
        return std(self[parm])

    def show(self):
        r"""
        Call *show()* from matplotlib to bring graphs to screen.
        """
        try:
            plt.show()
        except Exception as e:
            PE.warn(PE.PyAUnclassifiedError("Plot could not be shown. The following exception occurred:\n"
                                            + str(e)))

    def setBurn(self, burn):
        r"""
        Change value of "post burn-in".

        In the case of an emcee trace, the "post burn-in" is
        applied to the trace of all walkers.

        Parameters
        ----------
        burn : int
            The number of samples to be neglected.
        """
        self.burn = burn
        if self.dbtype == "emcee":
            # Apply burn-in to individual walkers
            self._loadEMCEEChain(burn=self.burn)

    def setThin(self, thin):
        r"""
        Change value of "post thinning".

        Parameters
        ----------
        thin : int
             Applies thinning to each chain.
             Retains every `k` th sample,
             where `k` is an integer value.

        Notes
        -----
        Use the "post thinning" to thin out your chains.
        """
        self.thin = thin

    def parameterSet(self, prescription="lowestDev"):
        r"""
        Find parameter values for a particular prescription.

        Parameters
        ----------
        prescription : string, {"lowestDev", "mean", "median"}
            Which parameter set to find. If 'lowestDev' is used,
            the method will return the parameter set pertaining to
            the lowest deviance solution. If 'mean' or 'median' are
            specified, the mean or median parameter values
            determined from the Markov Chains are returned.

        Returns
        -------
        Parameter set : dictionary
            A dictionary mapping all parameter names to the value
            derived using the specified prescription.
        Lowest deviance index : int
            The index of the lowest deviance solution. Only
            returned if the prescription is 'lowestDev'
        """
        if prescription == "lowestDev":
            result = {}
            indi = np.argmin(self["deviance"])
            for par in self.availableParameters():
                result[par] = self[par][indi]
            return result, indi
        elif prescription == "mean":
            result = {}
            for par in self.availableParameters():
                result[par] = self.mean(par)
            return result
        elif prescription == "median":
            result = {}
            for par in self.availableParameters():
                result[par] = self.median(par)
            return result
        else:
            raise (PE.PyAValError("Unknown state '" + str(prescription) + "'.",
                                  solution="Use either of 'lowestDev', 'mean', or 'median'.",
                                  where="parameterSet"))

    def setToState(self, model, state="best", verbose=True):
        r"""
        Set the parameter values to a certain state.

        Parameters
        ----------
        model - fitting object 
            The fitting model object whose parameters will be updated.
        state : {"best", "mean"}, optional
            "best" : Set parameters to the "best fit" state as measured
                     by deviance. This is the default.
            "mean" : Set parameters to mean value of trace.
        verbose : bool, optional
            If False, no output about what is done will be generated
            (default is True).
        """
        if verbose:
            print("Setting model to state: ", state)
        if state == "best":
            # Setting to best state as measured by deviance
            indi = np.argmin(self["deviance"])
            if verbose:
                print("Lowest deviance of ",
                      self["deviance"][indi], " at index ", indi)
            for par in self.availableParameters():
                if not par in six.iterkeys(model.parameters()):
                    continue
                model[par] = self[par][indi]
                if verbose:
                    print("Setting parameter: ", par,
                          " to value: ", model[par])
        if state == "mean":
            # Setting to "mean" state
            for par in self.availableParameters():
                if not par in six.iterkeys(model.parameters()):
                    continue
                model[par] = np.mean(self[par])
                if verbose:
                    print("Setting parameter: ", par,
                          " to mean value: ", model[par])



from __future__ import print_function, division
from PyAstronomy import funcFit as fuf
from PyAstronomy import pyasl
from PyAstronomy.pyaC import pyaErrors as PE
import numpy as np
import scipy.interpolate as sci
from PyAstronomy.modelSuite import ic
import six.moves as smo

class LLGauss(fuf.OneDFit):
  """
    A spectral model based on Gaussian profiles and a line list.
    
    This class provides a simple spectral model based on a number
    of Gaussian lines, whose strength may be fitted individually.
    Note that the EW of the lines is given by:
    `A{n}`*`lineScale`, where `A{n}` is the area of the n-th
    Gaussian. The `scale` parameter does not influence the
    EW of the Gaussians.
    
    .. note:: The unit of the EWs given in the `lineList` needs
              to be the same as the wavelength units.
    
    *Fit parameters*:
      - `lineScale` - A common scaling of the area of *all* lines.
      - `scale` - A scaling of the *entire* spectrum.
      - `eps` - Linear limb-darkening coefficient.
      - `vrad` - Radial velocity [km/s].
      - `vsini` - Projected stellar rotational velocity [km/s].
      - `A{n}` - The amplitudes (area) parameters of the individual Gaussians.
      - `sig{n}` - The standard deviations of the individual Gaussian.
      - `mu{n}` - The position of the individual Gaussians.
    
    Parameters
    ----------
    onlyAbs : boolean, optional
        If True (default), restrictions will be applied, which
        prevent emission lines in the spectrum.
    lineList : array
        An array with either two or three columns. The first column
        given the position of the lines, the second gives the EW
        of the lines, and the third---if present---gives the depth
        of the lines. The depth is the maximal depression of the
        continuum, e.g., a value of 0.96 means that the center
        of the line of 4% below the continuum. If the depth is given,
        the width of the individual Gaussians is determined from it,
        unless the `uniSig` is specified.
    uniSig : float, optional
        Use "unified sigma", i.e., the same width for all lines.
        Note that this flag overrules the "depth" column in the
        `lineList`, if it has been specified.
    modelBinsize : float, optional
        Internally, the model should be calculated on a finer grid
        than the actual spectrum. This parameter specifies the used
        bin size, which is 0.005 by default.
    useFastRB : boolean, optional
        Use the "fast" rotational broadening algorithm. This algorithm
        uses a wavelength-independent broadening kernel, which is
        considerably faster than considering the wavelength dependence.
        Setting this flag to False is necessary if you use very
        long wavelength ranges; by default it is True.
    verbose : boolean, optional
        If True, the class print the current parameters during the
        evaluation.
  """

  def __init__(self, lineList, uniSig=None, modelBinsize=0.005, useFastRB=True, verbose=False, onlyAbs=True):
    
    if not ic.check["scipy"]:
      raise(PE.PyARequiredImport("Could not import scipy.", \
                                 solution="Install scipy."))
    # Check whether depths are given
    self._depthsGiven = (len(lineList[0,::]) == 3)
    if (not self._depthsGiven) and (uniSig is None):
      raise(PE.PyAValError("No width and no depth given.", \
                           solution="Specify line depth in `lineList` or use `uniSig`."))
    # Check whether unified sigma shall be used
    # Note: Line depth will be ignored then
    self._uniSig = uniSig 
    
    # Only absorption lines
    self._onlyAbs = onlyAbs
    # Verbosity
    self._verbose = verbose
    # Use fast rotational broadening?
    self._usefastRB = useFastRB
    # Save binsize for the model
    self._modelBinsize = modelBinsize
    # Copy line list
    self._lineList = lineList.copy()
    # Number of lines
    self._nl = len(lineList[::,0])
    # A MultiGaussFit object
    self._mg = fuf.MultiGauss1d(self._nl)
    # Store all parameter names from GaussFit
    pars = list(self._mg.parameters().keys())
    # Remove lin and off from multiGauss keys
    pars.remove("lin")
    pars.remove("off")
    
    pars.extend(["lineScale", "scale", "eps", "vrad", "vsini"])
    fuf.OneDFit.__init__(self, pars)
    # Name the model
    self.setRootName("LineListGauss")
    
    # Assign start values
    for i in range(self._nl):
      p = str(i+1)
      # Assign position
      self._mg["mu"+p] = lineList[i,0]
      self["mu"+p] = lineList[i,0]
      # Assign amplitude/EW
      # Note: Positive EWs correspond to absorption lines
      self._mg["A"+p] = -lineList[i,1]
      self["A"+p] = lineList[i,1]
      # Assign width (sigma)
      if self._depthsGiven and (self._uniSig is None):
        # Depth IS given and no unified sigma specified
        self._mg["sig"+p] = abs(lineList[i,1])/((1.-lineList[i,2]) * np.sqrt(2.*np.pi))
        self["sig"+p] = self._mg["sig"+p] 
      elif not (self._uniSig is None):
        # uniSig IS given
        self._mg["sig"+p] = self._uniSig
        self["sig"+p] = self._mg["sig"+p] 
    
    self["scale"] = 1.0
    self["lineScale"] = 1.0
    
    if self._onlyAbs:
      # Apply restrictions to prevent emission lines
      for i in range(self._nl):
        p = str(i+1)
        self.setRestriction({"A"+p:[0.0, None]})

  def thawLineStrengths(self, wlmin=None, wlmax=None):
    """
      Thaw line strengths.
      
      Thaws parameters of the from A{n}, where
      n is the number of the Gaussian component.
      By default all such parameters will be
      thawed. The selection may, however, be
      influenced by specifying `wlmin` and `wlmax`.
      
      Parameters
      ----------
      wlmin : float, optional
          If specified, only the strength of lines at
          wavelengths larger than this limits will be thawed.
      wlmax : float, optional
          If specified, only the strength of lines at
          wavelengths below this limit will be thawed.
      
      Returns
      -------
      Thawed parameters : list
          A list of thawed parameter names.
    """
    freeAs = []
    for i in smo.range(self._nl):
      p = str(i+1)
      mu = self["mu"+p]
      if wlmin is not None:
        if mu < wlmin: continue
      if wlmax is not None:
        if mu >= wlmax: continue
      freeAs.append("A"+p)
      
    self.thaw(freeAs)
    return freeAs

  def thawLineWidths(self, wlmin=None, wlmax=None):
    """
      Thaw line widths.
      
      Thaws parameters of the from sig{n}, where
      n is the number of the Gaussian component.
      By default all such parameters will be
      thawed. The selection may, however, be
      influenced by specifying `wlmin` and `wlmax`.
      
      Parameters
      ----------
      wlmin : float, optional
          If specified, only the strength of lines at
          wavelengths larger than this limits will be thawed.
      wlmax : float, optional
          If specified, only the strength of lines at
          wavelengths below this limit will be thawed.
      
      Returns
      -------
      Thawed parameters : list
          A list of thawed parameter names.
    """
    freeSigs = []
    for i in smo.range(self._nl):
      p = str(i+1)
      mu = self["mu"+p]
      if wlmin is not None:
        if mu < wlmin: continue
      if wlmax is not None:
        if mu >= wlmax: continue
      freeSigs.append("sig"+p)
      
    self.thaw(freeSigs)
    return freeSigs
    
  def numberOfLines(self):
    """
      Get the number of lines in the model.
      
      Returns
      -------
      Number of lines : int
          Number of Gaussian in the model.
    """
    return self._nl

  def evaluate(self, x):
    """
      Calculates the model for current parameters.

      The "model" is calculated on a wavelength axis with
      binning specified by the `modelBinsize` parameter.
      
      The line positions are Doppler shifted and the resulting
      model is rotationally broadened. Finally, the entire
      model is multiplied by the `scale` parameter to
      account for a global normalization. 

      Parameters
      ----------
      x : array
        The wavelengths at which to calculate the model.
      
      Returns
      -------
      model : array
          The model evaluated at the specified positions.
    """
    if self._verbose:
      print("Evaluating with parameters: ", self.parameters())
    # Calculate how much longer the model wavelength axis should
    # be to ensure that Doppler shift and rotational broadening
    # to not "violate" its edges. 
    maxV = abs(self["vsini"]) + abs(self["vrad"])
    deltaWvl = max(x) * (1.0 + maxV/299792.458) - max(x)
    # The additional summand `_modelBinsize` is needed to ensure
    # that the wavelength axis is long enough even if the velocities
    # are very small. 
    mwvl = np.arange(min(x)-deltaWvl, max(x)+deltaWvl+self._modelBinsize, self._modelBinsize)
    
    mflux = np.ones(len(mwvl))
    # Set parameters of multiGauss
    for i in smo.range(self._nl):
      p = str(i+1)
      # Apply global scaling of amplitudes
      self._mg["A"+p] = -self["A"+p] * self["lineScale"]
      self._mg["sig"+p] = self["sig"+p]
      # Apply doppler shift of lines
      self._mg["mu"+p] = (1.0 + self["vrad"]/299792.458) * self["mu"+p]
    mflux += self._mg.evaluate(mwvl)
    
    if abs(self["vsini"]) > 0.0:
      if self._usefastRB:
        mflux = pyasl.fastRotBroad(mwvl, mflux, self["eps"], np.abs(self["vsini"]))
      else:
        mflux = pyasl.rotBroad(mwvl, mflux, self["eps"], np.abs(self["vsini"]))
    
    mflux *= self["scale"]
    
    # Return model interpolated at input wavelength
    f = sci.interp1d(mwvl, mflux)
    return f(x)



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import numpy as np
from .onedfit import OneDFit


class SinusFit1d(OneDFit):
    r"""
    Implements a sinusoidal function of the form:
       A * sin[ 2*pi * (nu * t + phi) ] + off.

    *Fit parameters*:
     - `A` - Amplitude
     - `nu` - Frequency (= 1/Period)
     - `phi` - Phase
     - `off` - Offset
    """

    def __init__(self):
        OneDFit.__init__(self, ["A", "nu", "phi", "off"])
        self.setRootName("Sine")

    def evaluate(self, x):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        x : array
            Specifies the points at which to evaluate the model.
        """
        y = self["A"]*np.sin(2.0*np.pi*(self["nu"]*x+self["phi"])) \
            + self["off"]
        return y


class ExpDecayFit1d(OneDFit):
    r"""
    Implements an exponential decay function of the form
       A * exp[ - (t-t0)/tau] + off.

    *Fit parameters*:
     - `A` - Amplitude
     - `tau` - Mean lifetime (=1/decay rate)
     - `t0` - Onset time
     - `off` - Continuum offset
    """

    def __init__(self):
        OneDFit.__init__(self, ["A", "tau", "t0", "off"])
        self.setRootName("ExpDecay")

    def evaluate(self, x):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        x : array
            Specifies the points at which to evaluate the model.
        """
        y = np.zeros(len(x)) + self["off"]
        indi = x > self["t0"]
        y[indi] += self["A"]*np.exp(-(x[indi]-self["t0"])/self["tau"])
        return y



# Copyright 2016 Stefan Czesla, Tanja Molle

# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the
# following conditions:

# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
# AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import print_function, division
import numpy as np
import six.moves as smo
import scipy.special as ss


class SampCorr(object):

    def __init__(self):
        pass

    def getak(self, N):
        # Calculate the required coefficients (a_k)
        ak = np.zeros(N + 2)
        for k in smo.range(N + 2):
            ak[k] = (-1.0)**k * ss.binom(N + 1, k)
        return ak

    def get_rhok(self, N):
        ak = self.getak(N)
        rho = np.zeros(len(ak))
        a = np.concatenate((ak, np.zeros(len(ak))))
        b = np.concatenate((ak, np.zeros(len(ak))))
        r0 = np.sum(a * b)
        for k in smo.range(len(ak)):
            rho[k] = np.sum(a * b) / r0
            b = np.roll(b, 1)
        return rho

    def get_r(self, N, ss=None):
        rk = self.get_rhok(N)
        if ss is None:
            # Sample size not given
            r = 1.0 + 2.0 * np.sum(rk[1:])
        else:
            # Sample size given
            k = np.arange(len(rk))
            r = 1.0 + 2.0 * np.sum((ss - k[1:]) / ss * rk[1:])
        return r

    def get_nb(self, N, ss):
        """
        BH46 Eq. 14
        """
        rk = self.get_rhok(N)
        # Sample size given

        onb = 1. / ss

        k = np.arange(len(rk), dtype=float)
        #print("k: ", k[1:], (1.-k[1:]/ss), rk )
        onb += (2.0 / ss) * np.sum((1. - k[1:] / ss) * rk[1:])
        nb = 1. / onb
        return nb, onb

    def get_nvr(self, N):
        rho = self.get_rhok(N)
        rf = np.sum(rho**2)
        return rf


class BSBase(object):
    """
    Functionality to estimate noise from equidistantly and arbitrarily sampled data.
    """

    def __init__(self):
        # Factor to convert MAD into std estimate (0.67448975019608171)
        self._madfac = np.sqrt(2.) * ss.erfinv(2. * 0.75 - 1.)

    def _checkJP(self, j):
        """
        Check validity of specified jump parameter.
        """
        if j < 1:
            raise(ValueError(
                "Jump parameter must be an integer equal or larger than 1. Currently, a value of " + str(j) + " is specified."))

    def _checkN(self, N):
        """
        Check validity of specified order of approximation.
        """
        if N < 0:
            raise(ValueError(
                "Order of approximation (N) must be an integer larger than zero. Currently, a value of " + str(N) + " is specified."))

    def subsetIndexDissection(self, ndp, N, j):
        """
        Find array indices of (N+2)-length subsets.

        Here, (N+2)-sized sets of indices are constructed, which
        are required in the construction of the beta sample.

        Parameters
        ----------
        ndp : int
            Number of available data points
        N : int
            Last order of the Taylor series taken into account. Chunk
            length will be N+2.
        j : int
            Jump parameter (>0)

        Returns
        -------
        k-indices : list
            A list holding N+2 elements. Each element of the list is an
            array holding the indices of the data point no. one, two, ..., N+2
            of the subsets required to construct the beta sample.
        """
        self._checkJP(j)

        # Range of available indices
        x = np.arange(ndp, dtype=int)

        # Chunk length from which to extract subsamples
        step = (2 + N) * j

        # Temporarily store indices of subsets
        tmp = [list() for _ in smo.range(N + 2)]

        for k in smo.range(N + 2):
            for cj in smo.range(j):
                # Loop over jump parameter (cj = current jump)

                # Collect all data points being first, second, etc. in
                # individual chunks

                # The middle expression achieves that only complete subsets are considered.
                # The "or None" statement effectively makes 0 indicate the entire index range.
                # From point with no. k, the distance to the last point is l(k) = (N+1)*j+1 - k*j.
                s = x[k * j + cj:-((N + 1) * j + 1 - k * j) + 1 or None:step]
                tmp[k].extend(s.copy())

        # Convert into arrays
        result = [np.array(x) for x in tmp]

        return result

    def subsetIndices(self, cd):
        """
        Compose indices sets defining the subsets to construct the beta sample.

        Parameters
        ----------
        cd : list
            The output of subsetIndexDissection

        Returns
        -------
        Subset indices : 2d array
            A 2d-array holding the indices of all subsets, arranged
            so that result[i,::] holds the N+2 indices pertaining
            to the i-th subset.
        """
        # Length of individual subsets (N+2)
        cl = len(cd)
        # Number of chunks
        nc = len(cd[0])

        result = np.zeros((nc, cl), dtype=int)
        for i in smo.range(cl):
            result[::, i] = cd[i]
        return result

    def stdc4(self, n):
        """
        Calculate c4 factor.

        The c4 factor is required to obtain an unbiased estimator
        for the standard deviation.

        It is proportional to the factor B used in Kenney 1940, who
        started from a slightly different definition of the sample
        variance.

        Parameters
        ----------
        n : int
            Number of points in the sample

        Returns
        -------
        c4, ln(c4) : float
            The c4 factor and its natural logarithm.
        """
        lnc4 = 0.5 * (np.log(2.0) - np.log(n - 1.0)) + \
            ss.gammaln(n / 2.0) - ss.gammaln((n - 1.) / 2.)
        c4 = np.exp(lnc4)
        return c4, lnc4

    def stdUnbiased(self, y):
        """
        Get unbiased estimate of the standard deviation and its standard deviation.

        Parameters
        ----------
        y : array
            Sample from which to estimate standard deviation.

        Returns
        -------
        Std : float
            Unbiased estimate of the standard deviation.
        Std of std : float
            Standard deviation of the unbiased standard deviation estimator.
        """
        s = np.std(y, ddof=1)
        c4 = self.stdc4(len(y))[0]
        return s / c4, s / c4 * np.sqrt(c4**(-2) - 1.0)

    def meanUnbStd(self, x):
        """
        Mean and (unbiased) standard deviation of the mean.

        Parameters
        ----------
        x : array
            Sampe from which to calculate mean and std

        Returns
        -------
        mean, std : floats
            Mean and unbiased standard deviation of the mean
        """
        m = np.mean(x)
        s = self.stdUnbiased(x)[0] / np.sqrt(len(x))
        return m, s

    def estimateStdMAD(self, x, mode):
        """
        Estimate standard deviation based on median absolute deviation (MAD)

        Parameters
        ----------
        x : array
            Sample from which to determine the standard deviation.
        mode : string, {zm, em}
            If 'zm', the population median is assumed to be zero.
            If 'em', the population median is estimated as the
            sample median.

        Returns
        -------
        std : float
            Estimate of the standard deviation
        """
        if mode == "zm":
            mad = np.median(np.abs(x))
        elif mode == "em":
            mad = np.median(np.abs(x - np.median(x)))
        sigest = mad / self._madfac
        return sigest

    def variance(self, x, mode):
        """
        Estimate variance from sample

        Parameters
        ----------
        x : array
            sample
        mode : string, {n, nmo}
            Estimator to use: 'n' for 1/n version with
            zero mean (not estimated) and 'nmo' for 1/(n-1)
        """
        if mode == "n":
            # 1/n version
            return np.sum(x**2) / len(x)
        elif mode == "nmo":
            # 1/(n-1) version
            m = np.mean(x)
            return np.sum((x - m)**2) / (len(x) - 1.)
        else:
            raise(ValueError("Unknown mode for variance: " + str(mode)))


class BSArbSamp(BSBase):
    """
    Estimate noise in equidistantly sampled data.
    """

    def __init__(self):
        super(BSArbSamp, self).__init__()
        self.betaSample = None
        self.betaSig, self.stdBetaSig = None, None
        self.mbeta, self.stdmbeta = None, None

    def getCoeffsArbSamp(self, t, gamma=1.):
        """
        Calculate coefficients (ak) for arbitrary sampling.

        Parameters
        ----------
        t : array
            Sampling instants of the subset.
        gamma : float, optional
            Scaling of the coefficients (default is one).

        Returns
        -------
        ak : array
            Set of coefficients.
        """
        np2 = len(t)
        Tp = np.matrix(np.zeros((np2 - 1, np2 - 1)))
        ti = np.zeros(np2 - 1)
        for i in smo.range(np2 - 1):
            Tp[i, ::] = t[1:]**i
            ti[i] = t[0]**i
        ap = -np.array(np.dot(Tp.I, np.matrix(ti).T)).ravel()
        a = np.ones(np2)
        a[1:] = ap
        return gamma * a

    def getBetaSampleShift(self, x, y, N, j):
        """
        Construct beta sample using shifting procedure.

        Parameters
        ----------  
        x, y : array
            Values from which to estimate noise
        N : int
            Last order of the Taylor series taken into account
        j : int
            Jump parameter (>0)

        Returns
        -------
        Beta sample : array
            An array holding all available beta values.
        """

        # Required chunk length to obtain realization
        # of beta = (N+2) + (N-1)*(j-1) = (N+1)*j + 1
        cl = (N + 1) * j + 1

        b = np.zeros(len(y) - cl + 1)

        for i in smo.range(len(y) - cl + 1):
            # Realizations of the beta sample
            ak = self.getCoeffsArbSamp(x[i:i + cl:j])
            b[i] = np.sum(y[i:i + cl:j] * ak)
            # Normalize to obtain width of sigma_0
            b[i] /= np.sqrt(np.sum(ak**2))

        return b

    def getBetaSample(self, x, y, N, j):
        """
        Combine data points to calculate beta values.

        Parameters
        ----------
        x : array
            Sampling of data
        y : array
            Values from which to estimate noise
        N : int
            Last order of the Taylor series taken into account
        j : int
            Jump parameter (>0)

        Returns
        -------
        Betas : array
            An array holding all available beta values.
        """
        self._checkJP(j)
        self._checkN(N)

        self.karr = self.subsetIndexDissection(len(y), N, j)
        # Indices of subsets required to construct beta sample
        ssindices = self.subsetIndices(self.karr)

        result = np.zeros(len(ssindices))

        for c in smo.range(len(ssindices)):
            chunkSamp = x[ssindices[c, ::]].copy()
            # Subtract first value for numerical reasons
            chunkSamp -= chunkSamp[0]
            ak = self.getCoeffsArbSamp(chunkSamp)

            for k in smo.range(N + 2):
                result[c] += ak[k] * y[ssindices[c, k]]

            # Normalize to obtain width of sigma_0
            result[c] /= np.sqrt(np.sum(ak**2))
        return result

    def betaSigma(self, x, y, N, j, ignoreNaN=True, returnMAD=False, ibs=False):
        """
        Estimate standard deviation of noise term in data set y.

        In this implementation, arbitrary sampling is taken into account.

        The method assigns the following attributes, which may be accessed after execution
        to work with the result:

        ============  ========  ===========================================
        Attribute     Type      Meaning
        ------------  --------  -------------------------------------------
        betaSample    array     The beta sample

        estimates     dict      Summary of the estimates obtained
                                from the beta sample (bs).
                                
                                "s2E": variance estimate of the bs
                                (expectation value of zero),
                                "s2Evar": variance of s2E
                                
                                "sE": Estimate of std of bs,
                                "sEstd": Std of sE
                                
                                "s2": variance estimate of the bs
                                (mean estimated from bs),
                                "s2var": variance of s2,
                                "s": sqrt(s2)
                                
                                "sME": Standard deviation based on MAD
                                with zero expectation,
                                "sMEstd": Estimation of std of sME
                                
                                "sM": Standard deviation based on MAD
                                with median estimated from bs,
                                "sMEstd": Estimation of std of sM
        ============  ========  ===========================================

        Parameters
        ----------   
        x : array
            Sampling of the data.
        y : array
            Data values from which to estimate standard deviation of noise.
        N : int
            Last order of the Taylor series to be taken into account.
        j : int, optional
            Jump parameter (default is one, i.e., consecutive data points are combined
            to estimate the noise).
        ignoreNaN : boolean, optional
            If True (default), NaN values in the beta sample are ignored in the
            calculation.
        returnMAD : boolean, optional
            If True, the estimate obtained using the MAD is returned instead of that
            of the MVU estimator (default is False). The standard error is estimated by
            scaling the standard deviation of the MVU estimator by a factor of
            1.64.
        ibs : boolean, optional
            If True, an independent beta sample is constructed. Default is False.

        Returns
        -------
        Estimate of STD in beta sample and the STD of the estimate: float, float
            The standard deviation determined in the beta sample. Non-robust
            estimates sE and sEstd if `returnMAD` is False (default) or robust
            estimates sME and sMEstd if `returnMAD` is True. The `estimates`
            attribute holds a more comprehensive summary of the estimates.
        """
        self._checkJP(j)
        self._checkN(N)

        if ibs:
            # Independent beta sample
            self.betaSample = self.getBetaSample(x, y, N, j)
        else:
            # Correlated beta sample
            self.betaSample = self.getBetaSampleShift(x, y, N, j)

        if len(self.betaSample) < 2:
            raise(ValueError(
                "Size of beta sample is too small (" + str(len(self.betaSample)) + ")"))

        if ignoreNaN:
            # Consider only non-NaN values
            bs = self.betaSample[np.isnan(self.betaSample) == False]
        else:
            # Use entire beta sample
            bs = self.betaSample

        # Get variance of beta sample
        self.bvar = self.variance(bs, mode="n")
        if ibs:
            # Independent beta sample
            self.bvarvar = 2. * self.bvar**2 / len(bs)
        else:
            # Dependent beta sample with correlation
            # Use estimate from independent sample to obtain estimate
            # because correlation structure is not exactly known
            self.bvarvar = 2. * self.bvar**2 / len(x) * (N + 2)

        # Estimate STD and its STD by propagation
        self.bstd = np.sqrt(self.bvar)
        self.bstdstd = np.sqrt(self.bvarvar / (4. * self.bvar))

        # Get standard deviation of beta sample and its standard error
        #self.betaSig, self.stdBetaSig = self.stdUnbiased(bs)
        # Get mean value of beta sample and its standard deviation
        self.mbeta, self.stdmbeta = self.meanUnbStd(bs)
        # Estimate the standard deviation based on MAD
        self.madstd = self.estimateStdMAD(bs, mode='zm')

        if ibs:
            self.stdMadstd = np.sqrt(
                1. / (2. * len(bs))) * self.estimateStdMAD(bs, mode='zm') * 1.64
        else:
            # Dependent beta sample with correlation
            # Use estimate assuming independent sample (for size of beta sample)
            # to obtain estimate of correlation
            lbs = len(x) / (N + 2)
            self.stdMadstd = np.sqrt(1. / (2. * lbs)) * \
                self.estimateStdMAD(bs, mode='zm') * 1.64

        # Estimates and variances
        # svar : Assumed to be equal to sEvar (more accurate formula given
        #        by Bayley and Hammersley 1946)
        self.estimates = {"s2E": self.bvar, "s2Evar": self.bvarvar,
                          "sE": self.bstd, "sEstd": self.bstdstd,
                          "s2": self.variance(bs, mode="nmo"),
                          "s2var": self.bstd,
                          "s": np.sqrt(self.variance(bs, mode="nmo")),
                          "sME": self.madstd, "sMEstd": self.stdMadstd,
                          "sM": self.estimateStdMAD(bs, mode='em'),
                          "sMEstd": self.stdMadstd}

        if not returnMAD:
            return self.bstd, self.bstdstd
        else:
            return self.madstd, self.stdMadstd


class BSEqSamp(BSBase):

    def __init__(self):
        super(BSEqSamp, self).__init__()
        self.betaSample = None
        self.betaSig, self.stdBetaSig = None, None
        self.mbeta, self.stdmbeta = None, None

    def get_ak(self, N):
        """
        Calculate the required coefficients (a_k)

        Parameters
        ----------
        N : int
            Order of approximation

        Returns
        -------
        ak : array
            The coefficients
        """
        ak = np.zeros(N + 2)
        for k in smo.range(N + 2):
            ak[k] = (-1.0)**k * ss.binom(N + 1, k)
        return ak

    def get_rhok(self, N):
        """
        Calculate (auto)correlation function for beta sample

        Parameters
        ----------
        N : int
            Order of approximation

        Returns
        -------
        correlation function : array
            Correlation function
        """
        ak = self.get_ak(N)
        rho = np.zeros(len(ak))
        a = np.concatenate((ak, np.zeros(len(ak))))
        b = np.concatenate((ak, np.zeros(len(ak))))
        r0 = np.sum(a * b)
        for k in smo.range(len(ak)):
            rho[k] = np.sum(a * b) / r0
            b = np.roll(b, 1)
        return rho

    def getBetaSample(self, y, N, j):
        """
        Construct the beta sample.

        Parameters
        ----------  
        y : array
            Values from which to estimate noise
        N : int
            Last order of the Taylor series taken into account
        j : int
            Jump parameter (>0)

        Returns
        -------
        Beta sample : array
            An array holding all available beta values.
        """
        self.karr = self.subsetIndexDissection(len(y), N, j)

        # Calculate the required coefficients (a_k)
        self.ak = self.get_ak(N)

        result = np.zeros(self.karr[0].size)

        for k in smo.range(N + 2):
            result += self.ak[k] * y[self.karr[k]]

        # Normalize to obtain width of sigma_0
        result /= np.sqrt(ss.binom(2 * N + 2, N + 1))
        return result

    def getBetaSampleShift(self, y, N, j):
        """
        Construct beta sample using shifting procedure.

        Parameters
        ----------  
        y : array
            Values from which to estimate noise
        N : int
            Last order of the Taylor series taken into account
        j : int
            Jump parameter (>0)

        Returns
        -------
        Beta sample : array
            An array holding all available beta values.
        """
        # Calculate the required coefficients (a_k)
        self.ak = self.get_ak(N)

        # Required chunk length to obtain realization
        # of beta = (N+2) + (N-1)*(j-1) = (N+1)*j + 1
        cl = (N + 1) * j + 1

        b = np.zeros(len(y) - cl + 1)

        for i in smo.range(len(y) - cl + 1):
            # Realizations of the beta sample
            b[i] = np.sum(y[i:i + cl:j] * self.ak)

        # Normalize to obtain width of sigma_0
        b /= np.sqrt(ss.binom(2 * N + 2, N + 1))
        return b

    def betaSigma(self, y, N, j, ignoreNaN=True, returnMAD=False, ibs=False):
        """
        Estimate standard deviation of noise term in data set y.

        It is explicitly assumed that the data are equidistantly sampled.

        ============  ========  ===========================================
        Attribute     Type      Meaning
        ------------  --------  -------------------------------------------
        betaSample    array     The beta sample

        estimates     dict      Summary of the estimates obtained
                                from the beta sample (bs).
                                
                                "s2E": variance estimate of the bs
                                (expectation value of zero),
                                "s2Evar": variance of s2E
                                
                                "sE": Estimate of std of bs,
                                "sEstd": Std of sE
                                
                                "s2": variance estimate of the bs
                                (mean estimated from bs),
                                "s2var": variance of s2,
                                "s": sqrt(s2)
                                
                                "sME": Standard deviation based on MAD
                                with zero expectation,
                                "sMEstd": Estimation of std of sME
                                
                                "sM": Standard deviation based on MAD
                                with median estimated from bs,
                                "sMEstd": Estimation of std of sM
        ============  ========  ===========================================

        Parameters
        ----------   
        y : array
            Data values from which to estimate standard deviation of noise.
        N : int
            Last order of the Taylor series to be taken into account.
        j : int
            Jump parameter
        ignoreNaN : boolean, optional
            If True (default), NaN values in the beta sample are ignored in the
            calculation.
        returnMAD : boolean, optional
            If True, the estimate obtained using the MAD is returned instead of that
            of the MVU estimator (default is False). The standard error is estimated by
            scaling the standard deviation of the MVU estimator by a factor of
            1.64.
        ibs : boolean, optional
            If True, an independent beta sample is constructed. Default is False.

        Returns
        -------
        Estimate of STD in beta sample and the STD of the estimate: float, float
            The standard deviation determined in the beta sample. Non-robust
            estimates sE and sEstd if `returnMAD` is False (default) or robust
            estimates sME and sMEstd if `returnMAD` is True. The `estimates`
            attribute holds a more comprehensive summary of the estimates.
        """
        self._checkJP(j)
        self._checkN(N)

        if ibs:
            # Get independent beta sample
            self.betaSample = self.getBetaSample(y, N, j)
        else:
            # Get correlated sample from shifting procedure
            self.betaSample = self.getBetaSampleShift(y, N, j)

        if len(self.betaSample) < 2:
            raise(ValueError(
                "Size of beta sample is too small (" + str(len(self.betaSample)) + ")"))

        if ignoreNaN:
            # Consider only non-NaN values
            bs = self.betaSample[np.isnan(self.betaSample) == False]
        else:
            # Use entire beta sample
            bs = self.betaSample

        # Get variance of beta sample
        self.bvar = self.variance(bs, mode="n")
        if ibs:
            # Independent beta sample
            self.bvarvar = 2. * self.bvar**2 / len(bs)
        else:
            # Dependent beta sample with correlation
            rhok = self.get_rhok(N)
            self.bvarvar = 2. * self.bvar**2 / \
                len(bs) * (1. + 2. * np.sum(rhok[1:]**2))

        # Estimate STD and its STD by propagation
        self.bstd = np.sqrt(self.bvar)
        self.bstdstd = np.sqrt(self.bvarvar / (4. * self.bvar))

        # Get standard deviation of beta sample and its standard error
        #self.betaSig, self.stdBetaSig = self.stdUnbiased(bs)
        # Get mean value of beta sample and its standard deviation
        self.mbeta, self.stdmbeta = self.meanUnbStd(bs)
        # Estimate the standard deviation based on MAD
        self.madstd = self.estimateStdMAD(bs, mode='zm')

        if ibs:
            rfac = 1.0
        else:
            # Dependent beta sample with correlation
            rhok = self.get_rhok(N)
            rfac = (1. + 2. * np.sum(rhok[1:]**2))
            # Estimate the STD of the estimate
        self.stdMadstd = np.sqrt(rfac / (2. * len(bs))) * \
            self.estimateStdMAD(bs, mode='zm') * 1.64

        # Estimates and variances
        # svar : Assumed to be equal to sEvar (more accurate formula given
        #        by Bayley and Hammersley 1946)
        self.estimates = {"s2E": self.bvar, "s2Evar": self.bvarvar,
                          "sE": self.bstd, "sEstd": self.bstdstd,
                          "s2": self.variance(bs, mode="nmo"),
                          "s2var": self.bstd,
                          "s": np.sqrt(self.variance(bs, mode="nmo")),
                          "sME": self.madstd, "sMEstd": self.stdMadstd,
                          "sM": self.estimateStdMAD(bs, mode='em'),
                          "sMEstd": self.stdMadstd}

        if not returnMAD:
            return self.bstd, self.bstdstd
        else:
            return self.madstd, self.stdMadstd




import numpy as np
import scipy.interpolate as sci
import gzip

import matplotlib
# matplotlib.use('TkAgg')
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
try:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2TkAgg
except ImportError:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk as NavigationToolbar2TkAgg
from matplotlib.figure import Figure

from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy import pyasl
from bisect import bisect

import six.moves.tkinter as tk
import six

from .pyaPicker import Point


class ContinuumInteractive:
    """
    GUI for interactive point selection.

    The `ContinuumInteractive` class provides a tool to normalize data
    interactively. The data are plotted and a number of points are defined
    using the mouse, which are connected by a spline to define a
    continuum estimate.

    Points are *selected* using the *middle button* of the mouse on the
    plot window. 

    Parameters
    ----------
    x, y : array
        The data to be normalized.
    config : dictionary, optional
        Information used for configuration:
          - specPlotStyle: Style used to plot the data (default 'b.--')
          - astyle: Style used to plot 'active' points (default 'ro')
          - istyle: Style used to plot 'inactive' points (default 'yp')
          - splineLineStyle: Style used to plot the continuum estimate (default 'r--')
          - normPlotRefLine: If True (default), a reference line is printed in the normalization window 
          - normLineStyle: Style used to plot the normalized data (default 'b.--')
          - normRefLineStyle: Style used to plot the reference line for the normalization (default 'k--')
          - sortPointListX: Determines whether point list is sorted in ascending order (default: True)
          - usegzip: If True (default), gzip is used to open and save files.
          - windowTitle: Title of the window

    Attributes
    ----------
    f : mpl Figure
        A Figure instance from matplotlib.
    a : mpl axis
        An Axis instance from matplotlib
    pointList : list of points
        A list containing the selected points in the form
        of `Point` instances.
    astyle : string
        Mpl style for plotting used for "active" point.
        Default is "ro".
    istyle : string
        Mpl style for plotting used for "inactive" point.
        Default is "yp".
    """

    def __init__(self, x, y, config=None):

        dconfig = {"specPlotStyle": "b.--",
                   "astyle": "ro",
                   "istyle": "yp",
                   "splineLineStyle": "r--",
                   "normLineStyle": "b.--",
                   "normPlotRefLine": True,
                   "normRefLineStyle": "k--",
                   "sortPointListX": True,
                   "usegzip": True,
                   "windowTitle": "PyA Continuum Interactive"}
        if config is not None:
            for k in six.iterkeys(dconfig):
                if k in config:
                    dconfig[k] = config[k]

        self.config = dconfig

        self.windowTitle = dconfig["windowTitle"]
        self.f = Figure()
        self.a = self.f.add_subplot(111)

        # The normalized plot figure
        self.normf = Figure()
        self.norma = self.normf.add_subplot(111)
        self._normaLineRef = None

        # Save the data (spectrum)
        self._x = x.copy()
        self._y = y.copy()
        self.a.plot(x, y, dconfig["specPlotStyle"])

        # Active and Inactive plot style for points
        self.astyle = dconfig["astyle"]
        self.istyle = dconfig["istyle"]

        self.pointList = []

        self.root = tk.Tk()

        # A frame containing the mpl plot
        self.plotFrame = tk.Frame()
        self.plotFrame.pack(fill=tk.BOTH, side=tk.LEFT, expand=True)
        self.canvas = FigureCanvasTkAgg(self.f, master=self.plotFrame)

        # A frame containing the box with selected points
        # and control buttons
        self.pointFrame = tk.Frame(self.root)
        self.pointFrame.pack(side=tk.LEFT, fill=tk.BOTH)

        self.listLabel = tk.Label(self.pointFrame, text="Selected points")
        self.listLabel.pack(side=tk.TOP)

        self.lb = tk.Listbox(self.pointFrame, height=15, selectmode=tk.SINGLE)
        self.lb.pack(side=tk.TOP)
        self.lb.bind("<<ListboxSelect>>", self._lbSelect)

        self.removeButton = tk.Button(master=self.pointFrame, text="Remove",
                                      command=self._removeButtonClicked)
        self.removeButton.pack(side=tk.TOP, fill=tk.X)

        self.clearButton = tk.Button(master=self.pointFrame, text="Clear all",
                                     command=self._clearAll)
        self.clearButton.pack(side=tk.TOP, fill=tk.X)

        # Says whether the normalized spectrum is currently shown
        self._normalizedDataShown = False
        # The dummy label only holds some space
        self.dummyLabel1 = tk.Label(self.pointFrame, text="")
        self.dummyLabel1.pack(side=tk.TOP)
        self.splineBoxLabel = tk.Label(self.pointFrame, text="Spline kind")
        self.splineBoxLabel.pack(side=tk.TOP)
        # Define spline selection
        self.splineOptions = ["Linear", "Quadratic", "INTEP"]
        self.splineFuncs = [self._evalLinearSpline,
                            self._evalQudraticSpline, self._evalIntepSpline]
        self.splineSelectBox = tk.Listbox(self.pointFrame, height=len(self.splineOptions), selectmode=tk.SINGLE,
                                          exportselection=False)
        for item in self.splineOptions:
            self.splineSelectBox.insert(tk.END, item)
        self.splineSelectBox.pack(side=tk.TOP)
        self.splineSelectBox.bind("<<ListboxSelect>>", self._splineSelected)
        self.splineSelectBox.select_set(0)
        self._splineSelected(None)

        self.normButton = tk.Button(master=self.pointFrame, text="Normalize",
                                    command=self._showNorm)
        self.normButton.pack(side=tk.TOP, fill=tk.X)

        # a tk.DrawingArea
        self.canvas.get_tk_widget().pack()
        self.cid = self.f.canvas.mpl_connect(
            'button_press_event', self._mouseButtonClicked)

        self.toolbar = NavigationToolbar2TkAgg(self.canvas, self.plotFrame)
        self.toolbar.update()
        self.canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        def _quit():
            if self._normalizedDataShown:
                self._quitWin()
            # stops main loop
            self.root.quit()
            # this is necessary on Windows to prevent
            # Fatal Python Error: PyEval_RestoreThread: NULL state
            self.root.destroy()

        self.quitButton = tk.Button(
            master=self.pointFrame, text='Quit', command=_quit)
        self.quitButton.pack(side=tk.BOTTOM, fill=tk.X)

        self.saveLoadFrame = tk.Frame(self.pointFrame)
        self.saveLoadFrame.pack(side=tk.BOTTOM, fill=tk.X)
        self.saveButton = tk.Button(
            master=self.saveLoadFrame, text="Save", command=self._saveToFile)
        self.saveButton.pack(side=tk.LEFT, fill=tk.X, expand=True)
        self.loadButton = tk.Button(
            master=self.saveLoadFrame, text="Load", command=self._loadFromFile)
        self.loadButton.pack(side=tk.RIGHT, fill=tk.X, expand=True)

        self.root.protocol("WM_DELETE_WINDOW", _quit)

    def _fileOpenMethod(self):
        """
        Get method for opening file.

        Determines whether gzip is used for file opening or not.

        Returns
        -------
        open : callable
            Method to open file.
        """
        if self.config["usegzip"]:
            return gzip.open
        return open

    def _pickleFileExtension(self):
        """
        Get default extension for pickle files.

        Returns
        -------
        Extension : string
            The default file extension.
        """
        if self.config["usegzip"]:
            return ".pickle.gz"
        else:
            return ".pickle"

    def _saveToFile(self, fn=None):
        """
        Save state to a pickle file.
        """
        import pickle
        pfe = self._pickleFileExtension()
        if fn is None:
            # Request a filename
            import six.moves.tkinter_tkfiledialog as tkFileDialog
            fn = tkFileDialog.asksaveasfilename(defaultextension=pfe, title="Save as pickle file",
                                                filetypes=[("pickle files", "*" + pfe)])
        pickle.dump(self._getState(), self._fileOpenMethod()(fn, 'w'))

    def saveStateToPickleFile(self, fn):
        """
        Save the state (point selection etc.) to pickle file.

        Parameters
        ----------
        fn : string
            The filename.
        """
        self._saveToFile(fn=fn)

    def _loadFromFile(self):
        """
        Load state from a pickle file.
        """
        import pickle
        import six.moves.tkinter_tkfiledialog as tkFileDialog
        pfe = self._pickleFileExtension()
        fn = tkFileDialog.askopenfilename(defaultextension=pfe, title="Load from pickle file",
                                          filetypes=[("pickle files", "*" + pfe)])
        state = pickle.load(self._fileOpenMethod()(fn))

        class Event:
            def __init__(self, xd, yd, inax):
                self.button = 2
                self.inaxes = inax
                self.xdata = xd
                self.ydata = yd

        for p in state["points"]:
            e = Event(p[0], p[1], self.a)
            self._mouseButtonClicked(e)

        self.splineSelectBox.select_clear(0, len(self.splineOptions))
        selected = self.splineOptions.index(state["splineKind"])
        self.splineSelectBox.select_set(selected)
        self._splineSelected(None)

    def _updateNormalizationPlot(self):
        """
        Replot the normalized spectrum.
        """
        if not self._normalizedDataShown:
            return

        normFlux = self._y / self._currentSpline

        if not self._normaLineRef is None:
            self.norma.lines.pop(self.norma.lines.index(self._normaLineRef))
        else:
            # First plot, add the reference line if requested
            if self.config["normPlotRefLine"]:
                self.norma.plot([self._x[0], self._x[-1]],
                                [1.0, 1.0], self.config["normRefLineStyle"])

        self.norma.plot(self._x, normFlux, self.config["normLineStyle"])
        self._normaLineRef = self.norma.lines[-1]
        self.normCanvas.draw()

    def _quitWin(self):
        self._normalizedDataShown = False
        if not self._normaLineRef is None:
            self.norma.lines.pop(self.norma.lines.index(self._normaLineRef))
        self._normaLineRef = None
        self._normwin.destroy()

    def _showNorm(self):
        """
        Shows normalized data in a separate window.
        """

        if self._normalizedDataShown:
            return

        self._normwin = tk.Tk()
        self._normwin.wm_title("Normalized spectrum")
        self._normwin.protocol("WM_DELETE_WINDOW", self._quitWin)

        # A frame containing the mpl plot
        self.normFrame = tk.Frame(master=self._normwin)
        self.normFrame.pack(fill=tk.BOTH, side=tk.LEFT, expand=True)
        self.normCanvas = FigureCanvasTkAgg(self.normf, master=self.normFrame)

        # a tk.DrawingArea
        self.normCanvas.get_tk_widget().pack()
        self.normToolbar = NavigationToolbar2TkAgg(
            self.normCanvas, self.normFrame)
        self.normToolbar.update()
        self.normCanvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        closeButton = tk.Button(master=self._normwin,
                                text='Close', command=self._quitWin)
        closeButton.pack(side=tk.BOTTOM)

        self._normalizedDataShown = True
        self._updateNormalizationPlot()

    def _pointListToArray(self):
        """
        Convert pointList data values into x,y arrays

        Returns
        -------
        x, y : array
            The x/y-data for the selected points. 
        """
        x, y = np.zeros(len(self.pointList)), np.zeros(len(self.pointList))
        for i, p in enumerate(self.pointList):
            x[i], y[i] = p.xdata, p.ydata
        indi = np.argsort(x)
        return x[indi], y[indi]

    def _evalLinearSpline(self):
        """
        Evaluates linear spline.

        Returns
        -------
        Spline : array
            The continuum estimate.
        """
        x, y = self._pointListToArray()
        if len(x) < 2:
            return self._x * np.nan
        f = sci.interp1d(x, y, bounds_error=False, fill_value=np.nan)
        return f(self._x)

    def _evalQudraticSpline(self):
        """
        Evaluates quadratic spline.

        Returns
        -------
        Spline : array
            The continuum estimate.
        """
        x, y = self._pointListToArray()
        if len(x) < 3:
            return self._x * np.nan
        f = sci.interp1d(x, y, bounds_error=False,
                         fill_value=np.nan, kind="quadratic")
        return f(self._x)

    def _evalIntepSpline(self):
        """
        Evaluates INTEP spline.

        Returns
        -------
        Spline : array
            The continuum estimate.
        """
        x, y = self._pointListToArray()
        if len(x) < 2:
            return self._x * np.nan
        return pyasl.intep(x, y, self._x, boundsError=False, fillValue=np.nan)

    def _splineSelected(self, event):
        """
        Called when the spline selection box is clicked.
        """
        selected = int(self.splineSelectBox.curselection()[0])
        self._evalSpline = self.splineFuncs[selected]
        self._evalSpline()
        self._updateSplineLine()
        self._updateView()
        # Save the name of the current spline selection
        self._currentSplineKind = self.splineOptions[selected]

    def _plotPointAs(self, state, point=None, lbString=None):
        """
        Plot a point in active/inactive color

        Parameters
        ----------
        state : string, {"active", "inactive"}
            Which color to use
        point : Point, optional
            The point information as an instance of the Point class.
        lbString : string
            The point identifier used in the listbox.
        """
        if (point is None) == (lbString is None):
            raise(PE.PyAValError("Either `point` or `lbString` have to be specified."))
        if point is not None:
            pli, lli = self._searchPoint(point.lbIdent)
        else:
            pli, lli = self._searchPoint(lbString)
        # Remove 'old' point from plot
        self.a.lines.pop(lli)
        if state == "active":
            style = self.astyle
        elif state == "inactive":
            style = self.istyle
        # Add new point in specified color
        self.a.plot([self.pointList[pli].xdata], [
                    self.pointList[pli].ydata], style)
        self.pointList[pli].mplLine = self.a.lines[-1]

    def _searchPoint(self, lbString):
        """
        Search point specified by listbox string. 

        Parameters
        ----------
        lbString : string
            The string used as identifier in the listbox.

        Returns
        -------
        Point list index : int
            The index of the point in the `pointlist`.
        Line index : int
            The index of the corresponding line in the
            lines list attribute of the axis instance. 
        """
        for ip, p in enumerate(self.pointList):
            if p.lbIdent == lbString:
                break
        for i, l in enumerate(self.a.lines):
            if l is p.mplLine:
                break
        return ip, i

    def _lbSelect(self, event):
        """
        React on change of point-listbox selection.

        This function determines the selected item
        and plots the point in "active" style.
        """
        # Get 'old' active item (changed afterward)
        oa = self.lb.get(tk.ACTIVE)
        # Get active item (string)
        ai = self.lb.get(int(self.lb.curselection()[0]))
        self._plotPointAs("inactive", lbString=oa)
        self._plotPointAs("active", lbString=ai)
        # Update view
        self._updateView()

    def _updateSplineLine(self):
        """
        Draws a new line representing the spline.
        """
        if not hasattr(self, "_splineLineRef"):
            self._splineLineRef = None
        if not self._splineLineRef is None:
            # Remove old line
            self.a.lines.pop(self.a.lines.index(self._splineLineRef))
        self._currentSpline = self._evalSpline()
        self.a.plot(self._x, self._currentSpline,
                    self.config["splineLineStyle"])
        self._splineLineRef = self.a.lines[-1]
        if self._normalizedDataShown:
            self._updateNormalizationPlot()

    def _mouseButtonClicked(self, event):
        """
        Called on click of mouse button.

        If the middle button has been clicked, a
        point is added to the selection.
        """
        # Accept only middle button
        if event.button != 2:
            return
        # Check whether click occurred on plot axes.
        if event.inaxes is not self.a:
            return

        p = Point(event)
        self.a.plot([p.xdata], [p.ydata], self.istyle)
        p.mplLine = self.a.lines[-1]

        index = tk.END
        if self.config["sortPointListX"]:
            # Insert point into the list at appropriate position
            # to keep x-values sorted
            x, y = self._pointListToArray()
            index = bisect(x, p.xdata)

        # Insert the point's content into the list box
        self.lb.insert(index, p.asStr())
        # Save the "list box identifier"
        p.lbIdent = p.asStr()
        # Save the point
        self.pointList.append(p)

        self._updateSplineLine()
        self._updateView()

    def _removeButtonClicked(self):
        """
        Remove a point.
        """
        # Check whether anything was selected
        sel = self.lb.curselection()
        if len(sel) == 0:
            return
        # Find indices for the point
        el = self.lb.get(int(sel[0]))
        pli, lli = self._searchPoint(el)
        # Remove it
        self.a.lines.pop(lli)
        self.pointList.pop(pli)
        self.lb.delete(sel[0])
        self._updateSplineLine()
        self._updateView()

    def _clearAll(self):
        """
        Remove all previously selected points.
        """
        np = self.lb.size()
        for _ in six.moves.range(np):
            self.lb.select_clear(0, self.lb.size())
            self.lb.select_set(0)
            self._removeButtonClicked()

    def _updateView(self):
        """
        Redraw MPL canvas
        """
        self.f.canvas.draw()

    def plot(self, *args, **kwargs):
        """
        Plot on interactive canvas.

        Accepts all arguments and keywords also accepted by
        matplotlib's `plot` method.
        """
        self.a.plot(*args, **kwargs)

    def plotNorm(self, *args, **kwargs):
        """
        Plot on normalization window.

        Accepts all arguments and keywords also accepted by
        matplotlib's `plot` method.
        """
        self.norma.plot(*args, **kwargs)

    def _getState(self):
        """
        Collect current state in a dictionary.

        Returns
        -------
        State : dictionary
            The following keys are defined:
              - points: A list of two-float tuples holding the
                        x,y location of the selected points.
              - continuum : Array holding the continuum estimate
                            at the given x-values.
              - splineKind : A string specifying the selected
                             spline option.
              - normalizedData : An array holding the normalized
                                 data.
        """
        result = {}
        result["points"] = []
        for p in self.pointList:
            result["points"].append((p.xdata, p.ydata))
        result["continuum"] = self._currentSpline.copy()
        result["normalizedData"] = self._y / result["continuum"]
        result["splineKind"] = self._currentSplineKind
        return result

    def findContinuum(self):
        """
        Interactively find the continuum estimate.

        Returns
        -------
        State : dictionary
            The following keys are defined:
              - points: A list of two-float tuples holding the
                        x,y location of the selected points.
              - continuum : Array holding the continuum estimate
                            at the given x-values.
              - splineKind : A string specifying the selected
                             spline option.
              - normalizedData : An array holding the normalized
                                 data.
        """
        self.root.wm_title(self.windowTitle)
        try:
            self.canvas.show()
        except AttributeError:
            self.canvas.draw()
        tk.mainloop()
        return self._getState()



from __future__ import absolute_import
from .pyaErrTemplate import PyaErrTemplate


class PyAOrderError(PyaErrTemplate):

    def __init__(self, what, **keys):
        """
        When to be raised?

        Whenever operations seem to be carried out in the
        wrong order.
        """
        PyaErrTemplate.__init__(self, what, "PyA Order Error", **keys)


class PyAParameterConflict(PyaErrTemplate):

    def __init__(self, what, **keys):
        """
        When to be raised?

        This exception should be raised when conflicting/mutually exclusive
        parameters are received.
        """
        PyaErrTemplate.__init__(self, what, "PyA Parameter Conflict", **keys)


class PyANetworkError(PyaErrTemplate):

    def __init__(self, what, **keys):
        """
        When to be raised?

        This exception should be raised when an action through the network
        has failed (e.g., a download).
        """
        PyaErrTemplate.__init__(self, what, "PyA Network Error", **keys)


class PyANotImplemented(PyaErrTemplate):

    def __init__(self, what, **keys):
        """
        When to be raised?

        This exception should be raised when the function/method called has not been implemented. \
        Such a situation often occurs when a member is to be implemented in a derived class \
        (abstract base class concept in c++).
        """
        PyaErrTemplate.__init__(self, what, "PyA Not Implemented", **keys)


class PyAImportFailure(PyaErrTemplate, ImportError):

    def __init__(self, what, **keys):
        """
        When to be raised?

        When an import fails. This may not be fatal.
        """
        PyaErrTemplate.__init__(self, what, "PyA import failure", **keys)


class PyARequiredImport(PyaErrTemplate, ImportError):

    def __init__(self, what, **keys):
        """
        When to be raised?

        If a definitely needed package (e.g., numpy) cannot be imported.
        """
        PyaErrTemplate.__init__(self, what, "PyA import error", **keys)


class PyAUnclassifiedError(PyaErrTemplate):

    def __init__(self, what, **keys):
        """
        When to be raised?

        If an error occurred that cannot or shall not be specified further. 
        """
        PyaErrTemplate.__init__(self, what, "PyA unclassified error", **keys)


class PyADownloadError(PyaErrTemplate):

    def __init__(self, what, **keys):
        """
        When to be raised?

        A download could not successfully be carried out.
        """
        PyaErrTemplate.__init__(self, what, "PyA download error", **keys)


class PyAFileError(PyaErrTemplate):

    def __init__(self, fn, mode, **keys):
        """
        When to be raised?

        Whenever a file-related error is detected.

        Parameters
        ----------
        fn : string
            Name of the file in question.
        mode : string, {no, ne}
            - "no": File could not be opened.
            - "ne": File does not exist
            - "ae": File already exists
        """
        if mode == "no":
            what = "File '" + str(fn) + "' could not be opened."
        elif mode == "ne":
            what = "File '" + str(fn) + "' does not exist."
        elif mode == "ae":
            what = "File '" + str(fn) + "' already exists."
        PyaErrTemplate.__init__(self, what, "PyA file error", **keys)


class PyAAlgorithmFailure(PyaErrTemplate):

    def __init__(self, what, **keys):
        """
        When to be raised?

        Whenever an algorithm fails in accomplishing its purpose.
        """
        PyaErrTemplate.__init__(self, what, "PyA algorithm failure", **keys)


class PyADeprecationError(PyaErrTemplate):

    def __init__(self, what, **keys):
        """
        When to be raised?

        Whenever a deprecated part of code is called.
        """
        PyaErrTemplate.__init__(self, what, "PyA deprecation note", **keys)



from __future__ import print_function, division
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.pyasl import _ic
import os
import six.moves as smo
import six


def write1dFitsSpec(fn, flux, wvl=None, waveParams=None, fluxErr=None, header=None, clobber=False, refFileName=None, refFileExt=0):
    """
    Write a 1d spectrum with equidistant binning to a fits file.

    Write a 1d-spectrum to a file. Wavelength axis and header keywords
    are related through the following expression:
    wvl = ((np.arange(N) + 1.0) - CRPIX1) * CDELT1 + CRVAL1,
    where CRPIX1, CDELT1, and CRVAL1 are the
    relevant header keywords.

    The function allows to specify an existing fits extension, from
    which the header will be cloned. Alternatively, an arbitrary,
    user-defined header may be given.

    Parameters
    ----------
    fn : string
        Filename
    flux : array
        Flux array
    wvl : array, optional
        Wavelength array. Either the wavelength array or the header
        keywords have to be provided (see `waveParams`). 
    waveParams : dict, optional
        Wavelength information required in the fits-header.
        Required keys are CDELT, CRVAL, CRPIX or (CDELT1, CRVAL1, CRPIX1).
    fluxErr : array, optional
        Flux errors. If given, the error will be stored in an additional extension.
    header : dict, optional
        Dictionary with header information to be transfered to the new file.
        Note that the wavelength information will be overwritten
        by the information given to this routine.
        If both, a reference file to clone the header from and the header parameters
        are given, the cloned header from the reference file will be overwritten and
        extended by the keywords specified here.
    refFileName : string, optional
        Clone header keywords from a reference file.
        Note that the wavelength information will be overwritten by the information
        by the information given to this routine.
    refFileExt : int, optional
        Reference-file extension to be used for cloning the header keywords.
        The default is 0.

    """

    reservedKeyWords = ["BITPIX", "SIMPLE", "NAXIS",
                        "NAXIS1", "CDELT1", "CRPIX1", "CRVAL1", "EXTEND"]

    if (not _ic.check["pyfits"]) and (not _ic.check["astropy.io.fits"]):
        raise(PE.PyARequiredImport("Could neither import module 'pyfits' and 'astropy.io.fits'.",
                                   where="write1dFitsSpec",
                                   solution="Install pyfits: http://www.stsci.edu/institute/software_hardware/pyfits"))

    if wvl is None and waveParams is None:
        raise(PE.PyAValError("The wavelength axis is not defined, i.e., neither \'wvl\' nor \'waveParams\' is specified.",
                             where="write1dFitsSpec",
                             solution="Provide wavelength array as \'wvl\' or wavelength information as \'waveParams\'."))

    if wvl is not None and waveParams is not None:
        raise(PE.PyAValError("You provided the wavelength axis as \'wvl\' AND wavelength information as \'waveParams\'. Don't know which to use.",
                             where="write1dFitsSpec",
                             solution="Provide only wavelength array via \'wvl\' or wavelength information via \'waveParams\'."))

    if wvl is not None:
        # Check whether the wavelength axis is equidistant
        dwl = wvl[1:] - wvl[0:-1]
        if (np.max(dwl) - np.min(dwl)) / np.max(dwl) > 1e-6:
            raise(PE.PyAValError("Wavelength axis seems not to be equidistant.",
                                 where="write1dFitsSpec",
                                 solution=["Check wavelength array.",
                                           "Consider passing wavelength information via `waveParams`."]))

    if os.path.isfile(fn) and not clobber:
        raise(PE.PyAFileError(fn, "ae",
                              where="write1dFitsSpec",
                              solution="File exists; set clobber=True to overwrite."))

    if _ic.check["pyfits"]:
        import pyfits
    else:
        import astropy.io.fits as pyfits

    # Put the flux into the output file.
    # Generate primary HDU
    hdu = pyfits.PrimaryHDU(flux)

    if refFileName:
        # If the is a reference file, its header will be used to populate
        # the header of the file to be written.
        if not os.path.isfile(refFileName):
            raise(PE.PyAFileError(str(refFileName), "ne",
                                  where="write1dFitsSpec",
                                  solution="Check file name."))
        ff = pyfits.open(refFileName)[refFileExt]

        for k in ff.header.keys():
            if k not in reservedKeyWords and k != "COMMENT":
                try:
                    if len(k) > 8:
                        hdu.header["HIERARCH " + k] = ff.header[k]
                    else:
                        hdu.header[k] = ff.header[k]
                except:
                    PE.warn(PE.PyAValError("Cannot write keyword <" +
                                           str(k) + "> with content " + str(ff.header[k])))

    if header is not None:
        # A use-defined header was specified
        for k in header.keys():
            hdu.header[k] = header[k]

    # Header keywords relevant for wavelength axis
    hk = {}
    # The only allowed type here
    hk["CTYPE1"] = "Linear"
    # If wavelength array is provided, create header keywords:
    if wvl is not None:
        hk["CRPIX1"] = 1
        hk["CRVAL1"] = wvl[0]
        hk["CDELT1"] = float(wvl[-1] - wvl[0]) / (len(wvl)-1)
    elif waveParams is not None:
        requiredKeys = ["CRPIX", "CRVAL", "CDELT"]
        # Counts whether all required keywords are provided
        count = 0
        for k in requiredKeys:
            subCount = 0
            for p in six.iterkeys(waveParams):
                if p.upper() == k or p.upper() == k+"1":
                    hk[k+"1"] = waveParams[p]
                    subCount += 1
            if subCount == 1:
                count += 1
            else:
                # This occurs if, e.g., both "CRVAL" and "CRVAL1" are present
                raise(PE.PyAValError("The wavelength parameters provided via \'waveParams\' contain ambiguous parameters. " +
                                     "You provided multiple values for " + k + ".",
                                     where="write1dFitsSpec",
                                     solution="Check content of \'waveParams\' so that only one value for " + k +
                                     " is provided (including lower/upper case and presence of a trailing \'1\'."))
        if count < 3:
            raise(PE.PyAValError("You provided an incomplete set of waveParams.",
                                  where="write1dFitsSpec",
                                  solution="Required keywords are CRPIX, CRVAL, and CDELT."))

    # (Over-)Write wavelength-related header information
    for k in hk.keys():
        hdu.header[k] = hk[k]

    if not fluxErr is None:
        # An error on the flux was given
        hdue = pyfits.ImageHDU(fluxErr)
        for k in hk.keys():
            # Add wavelength information to error header
            hdue.header[k] = hk[k]
        hdulist = pyfits.HDUList([hdu, hdue])
    else:
        hdulist = pyfits.HDUList([hdu])

    hdulist.writeto(fn, overwrite=clobber)



from __future__ import print_function, division
import unittest
import numpy
from PyAstronomy import funcFit as fuf
import six


class TestFuncFitSanity(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testsanity_parameterAssignment1(self):
        gf = fuf.GaussFit1d()
        origVars = ["A", "mu", "sig", "off", "lin"]
        vals = [1.0, 2.0, 3.0, 4.0, 5.0]
        for k, v in zip(origVars, vals):
            gf[k] = v
        for k, v in zip(origVars, vals):
            self.assertEqual(v, gf[k])
        gf.assignValue(dict(zip(origVars, numpy.zeros(len(origVars)))))
        self.assertEqual(numpy.sum(list(gf.parameters().values())), 0.0)

    def testsanity_combine1(self):
        gf = fuf.GaussFit1d()
        gff = gf + gf + gf
        for p in six.iterkeys(gff.parameters()):
            gff[p] = numpy.random.random()
            gff.thaw(p)
            gff.thaw([p, p])
            gff.freeze(p)
            gff.freeze([p, p])
            gff.setRestriction({p: [None, None]})
        for prop in ["A", "mu", "sig", "off", "lin"]:
            for c in [1, 2, 3]:
                gff[prop, "Gaussian", c] = numpy.random.random()
                s = (prop, "Gaussian", c)
                gff.thaw(s)
                gff.thaw([s, s])
                gff.freeze(s)
                gff.freeze([s, s])
                gff.setRestriction({s: [None, 10.0]})

    def testsanity_description(self):
        r"""
        Check sanity of 'description'
        """
        gf = fuf.GaussFit1d()
        gff = gf + gf
        self.assertEqual(
            gff.description(),
            "Gaussian(No. 1) + Gaussian(No. 2)",
            "Wrong description: " + gff.description(),
        )

    def testsanity_gaussfit1d_parameterization(self):
        """ Test parameterization of GaussFit1d """
        import numpy as np

        gf_hs = fuf.GaussFit1d(prm=("h", "sig"))
        gf_as = fuf.GaussFit1d(prm=("A", "sig"))
        gf_af = fuf.GaussFit1d(prm=("A", "FWHM"))
        gf_hf = fuf.GaussFit1d(prm=("h", "FWHM"))

        fsig = 2*np.sqrt(2*np.log(2))
        print(f"fsig = {fsig}")

        gfs = [gf_hs, gf_as, gf_af, gf_hf]

        gf_as["A"] = 1
        gf_as["sig"] = 1

        gf_af["A"] = 1
        gf_af["FWHM"] = fsig

        gf_hs["h"] = 1/np.sqrt(2*np.pi)
        gf_hs["sig"] = 1

        gf_hf["h"] = 1/np.sqrt(2*np.pi)
        gf_hf["FWHM"] = fsig

        x = np.linspace(-3, 3, 100)
        ys = []
        for g in gfs:
            ys.append(g.evaluate(x))

        for i in range(1, len(ys)):
            print(f"Checking y({i}) against 0")
            np.testing.assert_allclose(ys[i], ys[0], rtol=0, atol=1e-8)

    def testsanity_gaussfit1d_parameterization_fit(self):
        """ Test parameterization of GaussFit1d (fitting) """
        import numpy as np

        np.random.seed(1919)

        gf_hs = fuf.GaussFit1d(prm=("h", "sig"))
        gf_as = fuf.GaussFit1d(prm=("A", "sig"))
        gf_af = fuf.GaussFit1d(prm=("A", "FWHM"))
        gf_hf = fuf.GaussFit1d(prm=("h", "FWHM"))

        fsig = 2*np.sqrt(2*np.log(2))
        print(f"fsig = {fsig}")

        gfs = [gf_hs, gf_as, gf_af, gf_hf]

        gf_as["A"] = 1
        gf_as["sig"] = 1

        x = np.linspace(-3, 3, 100)
        y = gf_as.evaluate(x)
        y += np.random.normal(0, 0.001, len(x))

        gf_hf.thaw("h")
        gf_hf.thaw("FWHM")
        gf_hf.thaw("mu")

        gf_hf["h"] = 0.1
        gf_hf["FWHM"] = 2

        gf_hf.fit(x, y, minAlgo="spfmp")
        gf_hf.parameterSummary()

        np.testing.assert_almost_equal(
            gf_hf["h"], 1/np.sqrt(2*np.pi), decimal=3)
        np.testing.assert_almost_equal(gf_hf["FWHM"], fsig, decimal=3)
        np.testing.assert_almost_equal(gf_hf["mu"], 0, decimal=3)

    def testsanity_modeGrenander(self):
        """ Test sanity of mode Grenander """
        import numpy as np
        from PyAstronomy import funcFit as fuf

        np.random.seed(1919)
        x = np.random.normal(0.9, 2, 10000)

        k = 20
        for p in [1, 2, 3, 4, 5, 6, 7]:
            m = fuf.modeGrenander(x, k, p)
            print(p, m)
            self.assertAlmostEqual(
                m, 0.9, delta=0.1, msg="Problem with modeGrenander")


class MultiVoigtSanity(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testsanity_oneVsMulti(self):
        r"""
        Checking MultiVoigt1d vs. Voigt1d
        """
        from PyAstronomy import funcFit as fuf
        import numpy as np

        v1 = fuf.Voigt1d()
        vm = fuf.MultiVoigt1d(3)

        x = np.linspace(-10.0, 10.0, 200)

        ps = {
            "lin": 0.0067,
            "off": -12.753,
            "A": 9.0,
            "ad": 2.17,
            "al": 1.78,
            "mu": -0.771,
        }

        v1.assignValues(ps)

        vm["lin"] = v1["lin"]
        vm["off"] = v1["off"]

        for i in range(1, 4):
            num = str(i)
            for p in ["ad", "al", "mu"]:
                vm[p + num] = v1[p]
            vm["A" + num] = v1["A"] / 3.0

        self.assertAlmostEqual(
            np.max(np.abs(vm.evaluate(x) - v1.evaluate(x))),
            0.0,
            delta=1e-12,
            msg="MultiVoigt and Voigt1d deviate with 'amplitude separation'.",
        )

        v1 = fuf.Voigt1d()
        vm = fuf.MultiVoigt1d(3)

        x = np.linspace(-10.0, 10.0, 200)

        ps = {
            "lin": 0.0067,
            "off": -12.753,
            "A": 9.0,
            "ad": 2.17,
            "al": 1.78,
            "mu": -0.771,
        }

        v1.assignValues(ps)

        vm["lin"] = v1["lin"]
        vm["off"] = v1["off"]

        for i in range(1, 4):
            num = str(i)
            for p in ["ad", "al", "mu"]:
                vm[p + num] = v1[p]
            if i == 1:
                vm["A" + num] = v1["A"]

        self.assertAlmostEqual(
            np.max(np.abs(vm.evaluate(x) - v1.evaluate(x))),
            0.0,
            delta=1e-12,
            msg="MultiVoigt and Voigt1d deviate with one nonvanishing profile in multiVoigt.",
        )



# -*- coding: utf-8 -*-
import numpy
import PyAstronomy.funcFit as fuf
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.modelSuite.XTran import _ZList
from .occultquad_pya import OccultQuadPy

try:
    # Try import from local directory
    from . import occultquad
    _importOccultquad = True
except ImportError:
    # Try import from extension
    try:
        from PyAstronomy_ext.forTrans import occultquad
        _importOccultquad = True
    except ImportError:
        _importOccultquad = False

try:
    from . import occultnl
    _importOccultnl = True
except ImportError:
    try:
        from PyAstronomy_ext.forTrans import occultnl
        _importOccultnl = True
    except ImportError:
        _importOccultnl = False


class MandelAgolLC(_ZList, fuf.OneDFit):
    r"""
    Analytical transit light-curves using the formulae provided by Mandel & Agol 2002.

    .. note :: The computation of transit light curves
               is done using the external *occultquad* FORTRAN library.

               This library can be installed, e.g., via

               pip install PyAstronomy_ext

               It can also be compiled manually using SciPy's f2py
               wrapper (http://www.scipy.org/F2py). Simply go to the
               *forTrans* directory of the source distribution of PyAstronomy,
               then invoke

               f2py -c occultquad.pyf occultquad.f

               f2py -c occultnl.pyf occultnl.f

               If no FORTRAN implementation is available, a python re-implementation is used.
               Performance may be impacted.

    :Model parameters:

    The set of parameters specifying this model depends on: the type
    of orbit chosen (circular or keplerian) and the type of limb darkening
    chosen (quadratic or non-linear).

    More information on the Keplerian orbit can be found here: :ref:`keplerorbitpyasl`

    *Orbital model parameters (circular orbit)*:
      - `p` - Radius ratio between planet and star.
      - `a` - Semi-major axis of planetary orbit [stellar radii].
      - `i` - Inclination of orbit in degrees (90 deg is *edge on* view).
      - `T0` - Time offset of transit center.
      - `per` - Period of planetary orbit.
      - `b` - Describes the flux ratio between a stellar companion and the main star (default is 0).

    *Orbital model parameters (Keplerian orbit)*:
      - `p` - Radius ratio between planet and star.
      - `a` - Semi-major axis of planetary orbit [stellar radii].
      - `i` - Inclination of orbit in degrees (90 deg is *edge on* view).
      - `per` - Period of planetary orbit.
      - `b` - Describes the flux ratio between a stellar companion and the main star (default is 0).
      - `tau` - Time of periapsis passage.
      - `Omega` - Longitude of the ascending node [deg].
      - `w` - Argument of periapsis [deg]. Note that the longitude if periapsis is given by Omega+w.
      - `e` - Orbital eccentricity (0-1).
      - `sed` - Secondary eclipse depth 

    *Limb darkening parameters (quadratic)*:
      - `linLib` - Linear limb-darkening coefficient.
      - `quadLimb` - Quadratic limb-darkening coefficient.

    *Limb darkening parameters (non-linear)*:
      - `a1` - Non-Linear limb-darkening coefficient.
      - `a2` - Non-Linear limb-darkening coefficient.
      - `a3` - Non-Linear limb-darkening coefficient.
      - `a4` - Non-Linear limb-darkening coefficient.

    :Limb-darkening laws:

    The quadratic limb-darkening law is given by:

    .. math :: \frac{I(\mu)}{I(1)}= 1 - linLimb \times (1-\mu) - quadLimb \times (1-\mu)^2

    The non-linear limb-darkening law is given by:

    .. math :: \frac{I(\mu)}{I(1)}= 1 - \sum_{n=1}^{4}{a_n(1-\mu^{n/2})}

    :Modeling of secondary eclipse:

    If the parameter `sed` is not equal zero, a secondary eclipse (occultation) is added to the model. The
    secondary eclipse is modeled via a purely geometric eclipse of the stellar and planetary disk, which would
    then be behind the star (no limb darkening). The parameter `sed` specifies the depth of the secondary eclipse
    when the planetary disk is completely covered by the stellar disk, i.e., the depth may not be reached in a
    grazing configuration.

    .. warning::
        Time units have to be consistent.

    Parameters
    ----------
    orbit : string, {"circular", "keplerian"}, optional
        Determines whether a circular or full keplerian
        planetary orbit is used in the calculations. The
        default is a circular orbit.
    ld : string, {"quad", "nl"}
        Determines whether quadratic or non-linear
        limb darkening is used. The default is quadratic
        limb darkening.
    collCheck : boolean, optional
        If True (default), the model will check whether there
        is a physical collision between the star and the planet
        on evaluating the model and raises an exception when there
        is one.
    pyfo : boolean, optional
        Use python implementation of FORTRAN routines. Default is False;
        if FORTRAN routines are unavailable, this is the fallback option.
        Performance may be impacted.
    """

    def __init__(self, orbit="circular", ld="quad", collCheck=True, pyfo=False):
        if not orbit in ["circular", "keplerian"]:
            raise (PE.PyAValError("Invalid option for orbit: " + str(orbit),
                                  soltuion="Use either 'circular' or 'keplerian'."))
        if not ld in ["quad", "nl"]:
            raise (PE.PyAValError("Invalid option for orbit: " + str(ld),
                                  soltuion="Use either 'quad' or 'nl'."))
        _ZList.__init__(self, orbit, collCheck)

        self._oq = OccultQuadPy()
        if ((not _importOccultquad) or pyfo) and (ld == "quad"):
            self._oqcalc = self._oq.occultquad
        else:
            self._oqcalc = occultquad.occultquad

        if (not _importOccultnl) and (ld == "nl"):
            raise (PE.PyARequiredImport("Could not import required shared object library 'occultnl.so'",
                                        solution=["Use 'pip install PyAstronomy_ext' to get it.",
                                                  "Invoke PyA's install script (setup.py) with the --with-ext option.",
                                                  "Go to 'forTrans' directory of PyAstronomy and invoke\n    f2py -c occultnl.pyf occultnl.f"]
                                        ))

        if orbit == "circular":
            plist = ["p", "a", "i", "T0", "per", "b", "sed"]
        else:
            # It is an elliptical orbit
            plist = ["p", "a", "i", "per", "b",
                     "e", "Omega", "tau", "w", "sed"]

        if ld == "quad":
            plist.extend(["linLimb", "quadLimb"])
        else:
            # Non-linear LD
            plist.extend(["a1", "a2", "a3", "a4"])

        fuf.OneDFit.__init__(self, plist)
        self.freeze(plist)
        self.setRootName("Occultquad")

        if orbit == "keplerian":
            self["w"] = -90.0

        self._orbit = orbit
        self._ld = ld

    def backendStatus(self):
        r"""
        Print information on the code being used for evaluation
        """
        print("Backend status for MandelAgolLC")
        print("    Quadratic limb-darkening (quad)")
        if self._oqcalc == self._oq.occultquad:
            print("        Using Python reimplementation of FORTRAN routines")
            print(" "*8 + "To install compiled FORTRAN code you may attempt to:")
            print(" "*10 + "- Use 'pip install PyAstronomy_ext' to get it")
            print(
                " "*10 + "- Invoke PyA's install script (setup.py) with the --with-ext option")
            print(" "*10 + "- Go to 'forTrans' directory of PyAstronomy and invoke 'f2py -c occultquad.pyf occultquad.f'")
        else:
            print("        Using FORTRAN implementation")

    def evaluate(self, time):
        r"""
        Calculate a light curve according to the analytical models
        given by Mandel and Agol.

        Parameters
        ----------
        time : array
            An array of time points at which the light curve
            shall be calculated.

        Returns
        -------
        Model : array
            The analytical light curve is stored in the property `lightcurve`.
        """

        # Translate the given parameters into an orbit and, finally,
        # into a projected, normalized distance (z-parameter)
        if self._orbit == "circular":
            self._calcZList(time - self["T0"])
        else:
            # The orbit is keplerian
            self._calcZList(time)

        # Use occultquad Fortran library to compute flux decrease
        df = numpy.zeros(len(time))
        if len(self._intrans) > 0:
            if self._ld == "quad":
                # Use occultquad Fortran library to compute flux decrease
                result = self._oqcalc(self._zlist[self._intrans], self["linLimb"], self["quadLimb"],
                                      self["p"], len(self._intrans))
            else:
                result = occultnl.occultnl(self["p"], self["a1"], self["a2"], self["a3"],
                                           self["a4"], self._zlist[self._intrans])
            df[self._intrans] = (1.0 - result[0])
            if self["sed"] != 0:
                df[self._inocc] = (1.0 - occultquad.occultquad(self._zlist[self._inocc], 0, 0,
                                                               self["p"], len(self._inocc))[0])/self["p"]**2*self["sed"]

        self.lightcurve = (1. - df) * 1. / \
            (1. + self["b"]) + self["b"] / (1.0 + self["b"])

        return self.lightcurve


MandelAgolLC_Rebin = fuf.turnIntoRebin(MandelAgolLC)



# -*- coding: utf-8 -*-
from PyAstronomy.pyaC import pyaErrors as PE
import numpy as np
import copy


def binningx0dt(
    x,
    y,
    yerr=None,
    x0=None,
    dt=None,
    nbins=None,
    reduceBy=None,
    removeEmpty=True,
    removeNoError=False,
    useBinCenter=True,
    useMeanX=False,
    nanHandling=None,
    yvalFunc=np.mean,
):
    """
    A simple binning algorithm.

    This algorithm uses a fixed bin-width to produce a binned
    data set. Either the bin-width, `dt`, or the number of bins,
    `nbins`, must be specified. The number of output bins may
    also depend on other flags such as, for example, `removeNoError`.

    If no errors are specified via `yerr`, the errors for the binned
    data are estimated as the standard deviation of the input data
    points divided by the square root of their number. If `yerr` has
    been specified, error propagation is used to determine the error.

    The behavior of the x-axis can be controlled via the
    `useBinCenter` flag.

    Values which cannot be determined will be indicated by NaN.
    Various flags can be used to remove such bins from the binned
    data set.

    Parameters
    ----------
    x, y : array
        The x and y data values.
    yerr : array, optional
        Errors on the data values.
    x0 : float, optional
        Starting time of first bin.
        Default is lowest given x value.
    dt : float, optional
        Width of a bin (either `dt`, `nbins` or `reduceBy` must
        be given).
    nbins : int, optional
        Number of bins to use (either `dt`, `nbins` or `reduceBy` must
        be given). Note that this specifies the number of bins into which
        the range from `x0` to the last data point is subdivided.
    reduceBy : int, optional
        Reduce the number of elements in the array by the given factor
        (either `dt`, `nbins` or `reduceBy` must be given). Note that
        in this case, `x0` is set to the first (minimum x-value) and
        the number of bins, n, is calculated according to the
        prescription: :math:`n = int(round(len(x)/reduceBy))`
    removeEmpty : boolean, optional
        If True (default), bins with no data points will be
        removed from the result.
    removeNoError : boolean, optional
        If True, bins for which no error can be determined
        will be removed from the result. Default is False.
    useBinCenter : boolean, optional
        If True (default), the time axis will refer to the
        center of the bins. Otherwise the numbers refer to
        the start of the bins.
    useMeanX : boolean, optional
        If True, the binned x-values refer to the mean x-value
        of all points in that bin.
        Therefore, the new time axis does not have to be equidistant.
    yvalFunc : callable, optional
        Function used to determine the value in a bin based on input data.
        Default is the mean value (np.mean). An alternative choice could, e.g.,
        be np.median.
    nanHandling : None, "ignore", float, (optional)
        Controls how NaNs in the data are handled.
          - None: By default (None),
            nothing is done and NaNs are treated as if they were valid
            input data, so that they are carried over into the binned data.
            This means that output bins containing NaN(s) will also end up
            as NaN(s). If 'ignore'
          - 'ignore': In this case, NaNs contained in the input data are
            removed from the data prior binning. Note however, that `x0`,
            unless specified explicitly, will still refer to the first data
            point, whether or not this holds a NaN value.
          - float: If a float is given, input data values containing NaNs
            are replaced by the given float before binning. Note that no error on
            the data (yerr) can be considered in this case, to avoid
            erronous treatment of un- or misspecified error values.

    Returns
    -------
    Binned data set : array
        An array with four columns: 1) The new x-axis,
        2) The binned data (the mean value of the data points located
        in the individual bins), 3) Error of binned data, 4) The
        number of input data points used to create the bin.
        For instance, the new x-values can be accessed
        using result[::,0].
    dt : float
        The width of the bins.
    """
    if len(x) != len(y):
        raise (PE.PyAValError("x and y need to have the same length."))
    if ((not dt is None) + (not nbins is None) + (not reduceBy is None)) != 1:
        raise (PE.PyAParameterConflict("Specify one of `dt`, `nbins`, or `reduceBy`."))
    if ((not x0 is None) + (not reduceBy is None)) > 1:
        raise (PE.PyAParameterConflict("Specify either `x0` or `reduceBy`."))
    if np.any(np.isfinite(x) == False):
        raise (
            PE.PyAValError(
                "X axis contains invalid values (NaN or inf).",
                solution="Remove invalid values from input.",
            )
        )
    if x0 is None:
        # Use smallest time as starting point
        x0 = np.min(x)
    if x0 > np.max(x):
        raise (
            PE.PyAValError(
                "The starting point, `x0`, is larger than the end time of the data.",
                solution="Use a smaller value.",
            )
        )

    # Use arrays in calculation. Only copy if conversion to numpy array
    # is required
    xl, yl, yerrl = False, False, False
    if not isinstance(x, np.ndarray):
        x = np.array(x)
        xl = True
    if not isinstance(y, np.ndarray):
        y = np.array(y)
        yl = True
    if not yerr is None:
        if not isinstance(yerr, np.ndarray):
            yerr = np.array(yerr)
            yerrl = True

    # nanHandling
    if nanHandling is not None:
        # As some manipulation of the data may be required, generate
        # local copies, unless this has already happened.
        if not xl:
            x = copy.copy(x)
        if not yl:
            y = copy.copy(y)
        if not yerrl:
            yerr = copy.copy(yerr)

        if nanHandling == "ignore":  # Remove bins containing NaN from the input data
            gi = np.isfinite(y)
            x, y = x[gi], y[gi]
            if not yerr is None:
                # Check if bins containing NaN from the input data
                yerr = yerr[gi]
        else:
            try:
                nanValue = float(nanHandling)
            except ValueError as ve:
                raise (
                    PE.PyAValError(
                        "Invalid value given for 'nanHandling'. Error was: " + str(ve),
                        solution="Use None, 'ignore', or float number.",
                    )
                )

            gi = np.isnan(y)
            y[gi] = nanValue
            if yerr is not None:
                raise (
                    PE.PyAValError(
                        "yerr is not permitted for option 'nanHandling=float'.",
                        solution="Remove yerr=... from the call.",
                    )
                )

    # Calculate the new number of array elements.
    if reduceBy is not None:
        nbins = int(round(len(x) / float(reduceBy)))
        if nbins == 0:
            nbins = 1  # Prevent empty return arrays
    if nbins is not None:
        # Use a specified number of bins.
        # Calculate bin length
        dt = (np.max(x) - x0) / float(nbins)
    # Start calculation
    # In which bin do the individual data points belong?
    inWhichBin = np.floor(((x - x0) / dt)).astype(int)
    # Lonely last bin correction
    # Brings the last data point into the last valid bin
    # instead of creating a new bin with that data point\
    # at its very beginning
    if nbins is not None:
        inWhichBin[np.where(inWhichBin == nbins)[0]] -= 1
    # Get the number of bins (start at x0 even if the
    # first bins do not contain any data points)
    nbins = np.max(inWhichBin) + 1
    # Bins with data
    bwd = np.unique(inWhichBin)
    # Sort data into the bins
    # Create output array (time, flux, error, data-point-counter)
    result = np.empty((nbins, 4))
    result[:] = np.nan
    # Assign time axis (beginning of bins)
    result[::, 0] = x0 + np.arange(nbins) * dt
    if useBinCenter:
        # Use the center of the bin for timing
        result[::, 0] += 0.5 * dt
    # Set data point counter (points/bin) to zero
    result[::, 3] = 0
    for b in bwd:
        indi = np.where(inWhichBin == b)[0]
        result[b, 3] = len(indi)
        result[b, 1] = yvalFunc(y[indi])
        if useMeanX:
            # Overwrite the time axis using the mean x-value
            result[b, 0] = np.mean(x[indi])
        if yerr is None:
            # No errors on data points are given
            if len(indi) > 1:
                result[b, 2] = np.std(y[indi]) / np.sqrt(result[b, 3])
            else:
                # No error if there is only a single point in the bin
                result[b, 2] = np.nan
        else:
            # There are errors on the data points
            # Use error propagation
            result[b, 2] = np.sqrt(np.sum(yerr[indi] ** 2)) / result[b, 3]

    if removeEmpty:
        # Remove bins without data points in it
        indi = np.where(np.invert(np.isnan(result[::, 1])))[0]
        result = result[indi, ::]
    if removeNoError:
        # Remove bins for which no error can be given
        indi = np.where(np.invert(np.isnan(result[::, 2])))[0]
        result = result[indi, ::]
    return result, dt



from __future__ import print_function, division
import six.moves as smo
from . import pyaErrors as PE
import numpy as np

def matrix2doutput(m, oformat="% 6.2e", colsep=" | ", rowNames=None, colNames=None, transpose=False, toScreen=True):
  """
    Format a matrix in readable form and write it to screen.
    
    The column is specified by the second index, e.g., the first
    entry in the second column is given by m[0,1]. The first
    entry in the third row is, consequently, given by m[2,0].
    
    Parameters
    ----------
    m : 2-dimensional array
        The data to be formatted.
    oformat : string or list of strings, optional
        The output format. If string, the same format string will
        be used for all columns. If a list of strings is given,
        the associated specifier will be used for each individual
        column.
    colsep : string, optional
        The separator used between columns.
    rowNames : list of strings, optional
        The names of the rows.
    colNames : list of strings optional
        The names of the columns.
    transpose : boolean, optional
        If True, the input matrix will be transposed. In effect,
        this exchanges the roles of columns and rows. Note, however,
        that the role of `colNames` and `rowNames` remains
        unaltered. The default is False.
    toScreen : boolean, optional
        If True (default), the result will be written to screen. 
    
    Returns
    -------
    Formatted matrix : list of strings
        The formatted output is a list of strings, which
        might be written to screen.
  """
  if transpose:
    m = np.transpose(m)
  
  ncol = len(m[0,::])
  nrow = len(m[::,0])
  if rowNames is not None:
    if nrow != len(rowNames):
      raise(PE.PyAValError("Number of row names does not match the size of the matrix.", \
                           where="matrix2doutput", \
                           solution="You may try to use `transpose=True'."))
  if colNames is not None:
    if ncol != len(colNames):
      raise(PE.PyAValError("Number of column names does not match the size of the matrix.", \
                           where="matrix2doutput",\
                           solution="You may try to use `transpose=True'."))
  
  # Check whether format is a list (one format for each column)
  if isinstance(oformat, list):
    pass
  else:
    oformat = [oformat] * ncol
  
  if len(oformat) != ncol:
    raise(PE.PyAValError("The number of format specifiers does not match the number of columns.", \
                         where="matrix2doutput",\
                         solution=["Adjust the number of specifiers.", \
                                   "You may try to use `transpose=True'."]))                          
  
  # Find width of columns
  if colNames is not None:
    colWidth = [len(x) for x in colNames]
  else:
    colWidth = [0] * ncol
  if rowNames is not None:
    # Width of "zeros" column with row names
    colWidth.insert(0, max([len(x) for x in rowNames]))
  else:
    # Use -1, if there are no row names
    colWidth.insert(0, -1) 
  # Determine width of rest of columns
  for i in smo.range(ncol):
    l = max([len(oformat[i] % x) for x in m[::,i]])
    colWidth[i+1] = max(colWidth[i+1], l)
  
  lines = []
  
  if colNames is not None:
    # Create header
    if colWidth[0] != -1:
      h = " " * colWidth[0] + colsep
    else:
      h = ""
    for i, n in enumerate(colNames):
      h += ("%" + str(colWidth[i+1]) + "s") % n
      if i != (len(colNames) - 1):
        h += colsep
    lines.append(h)
    lines.append("-" * len(h))
  # Add the data
  for i in smo.range(nrow):
    l = ""
    if rowNames is not None:
      # First, add the name of the row
      l += ("%" + str(colWidth[0]) + "s") % rowNames[i]
      l += colsep
    # Loop over the column indices
    for j in smo.range(ncol - 1):
      l += ("%" + str(colWidth[j+1]) + "s") % (oformat[j] % m[i, j]) + colsep
    # Add last one
    l += ("%" + str(colWidth[-1]) + "s") % ((oformat[-1] % m[i, -1]))
    lines.append(l)
  
  if colNames is not None:
    lines.append(lines[1])
  
  if toScreen:
    for l in lines:
      print(l)
  return lines

  